# Introduction {#mainpage}
This library is based on the older *aircraftGeometry* library and extends it to be more modular.
The modularity and flexibility is achieved by using the high performance [Computational Geometry Algorithms Library](https://www.cgal.org/) also known as **CGAL**.

The library tries to model the shapes needed for aircraft design as flexibel as possible and tries to limit the assumptions about the shapes as far as possible.
It does not assume a specific orientation of components in the 3D space, but models every component in its own local coordinate space which then can be transformed to any parent coordinate system.
This flexibility is great to very freely define complex shapes which have an arbitrary location/orientation in the 3D space, but is makes it a bit harder to use since you have to remember in which coordinate space you are currently working in.
This documentation explains the different coordinate systems and how they are derived from each other in a later section.

Each geometry object in this library is referred to as an entity and derives from Entity3D.
This class defines the general properties of the location and orientation of the geometry.
It is used to determine the transformation in between the different coordinate spaces.

Before we dive into the coordinate systems however, let's get to know how the actual geometry and shapes are represented using **CGAL**.

@attention &rarr; This library does only support discrete shapes. Meaning the shapes are **always** polygons where the discrete vertices of the polygon are connected with lines which form the edges of the polygon.

@note The Python binding of this library does __not__ include a complete implementation of **CGAL**!
If you want to use the full flexibility of **CGAL** you need to implement your tool in C++.

---

## 2D Geometry 
The base of every geometry are 2D sections. The base properties of such a shape are shown here:

![Section 2D](section2d.svg){html: width=30%}

A section **always** defines its local axes as `X` and `Y`. Where `X` is the *horizontal* axis and `Y` the *vertical* axis.
The origin point of the section is the origin point of the local coordinate system.
The local coordinate system in the scope of a section always refers to the shown *X* and *Y* axes.

The shape of the section consists of vertices which form the outline polygon. The polygon can have an arbitrary shape. It only needs to be simple, meaning no edges connecting the vertices should intersect each other.
A concrete implementation of this described polygon concept which is used in the library is the [Polygon_2](https://doc.cgal.org/latest/Polygon/index.html#Chapter_2D_Polygon) class from **CGAL**.

The concept of the polygon shape is formalized with the *C++* concept [Shape](@ref geom2.Shape).
You can use any shape class which satisfies this concept to create the surfaces.
This library currently implements two main types:
- [PolygonSection](@ref geom2.PolygonSection): For creating general "tube" based surfaces, i.e. fuselages or nacelles.
- [AirfoilSection](@ref geom2.AirfoilSection): For creating aerodynamic surfaces which have an airfoil shape.

Those two section types mainly differ via the applied terminology how their geometric parameters are called.
For example: An **AirfoilSection** can only set the *chord length* as opposed to a **PolygonSection** where you can set the *width* and *height* of the section independently.
&rarr; Refer to the documentation of each section type to see what parameters you can set.

---

## 3D Geometry
The connection from the 2D section to create 3D surfaces are [multi-section surfaces](@ref geom2.MultisectionSurface). A multi-section surface consists of multiple sections which form the single segments of the the surface. A projected view of such a surface might look like this:

![View of multi-section surface](surface_top.svg){html: width=30%}

The example consists of three surface sections which result in **two** surface segments.
Note, that the library rather uses sections and not segments and therefore does not allow for disconnected steps in between the sections directly. You can however insert two section at the same location, which have different shapes. This effectively gives a stepped shape when viewed from any side, but the data structure itself is still connected.

The main extrusion direction of the surface is along the local `Z` axis.
The origin points of each section are 3D coordinates and define the location of the section within the surface coordinate space. By moving the origin point, the complete section gets moved as well.

@note The order how you insert the sections in the surface **does** matter as it defines how the sections are connected.

The surfaces themselves have an origin point which defines their location in the 3D space of some parent entity.

The multi-section surface is basically just a `std::vector` which contains multiple section objects.
You can use this vector as a regular iterator and apply **STL** algorithms as you are used to.

---

## Coordinate Systems
You should have a basic idea of how the geometry is build up.
Now, comes the part where we position the geometry within one space and actually build up a complete aircraft in its final shape.

> Keep your right hand ready to always do the "three finger dance" to visualize the right hand coordinate space. Also be aware, **Euler angles** will be involved! :-)

The library currently differentiates three coordinate spaces:
1. `2D` : The local coordinate system of a 2D section profile/shape.
2. `3D` : The local coordinate system of a multi-section surface which can include multiple 2D sections.
3. `3D` : The global body-fixed coordinate system of the aircraft where all the surfaces finally live and are positioned.

Whenever one entity is contained within another parent entity, the **parent coordinate system** refers to this parent entity.

### Transform from 2D to 3D
Let's look at a simple example, which mainly uses 2D coordinates:

![Example coordinate transform](coordinate_example.png){html: width=30%}

The *local point* `p` has 2D coordinates and is defined within the *local coordinate* system of `child`. The `child` represents a 2D section here, so it technically does not have a `Z` axis in its local coordinate space. It is just shown here for completeness to always remember the right-hand convention.

The `child` entity is then contained within the 3D entity `parent`. The `child` has an offset within the `parent` coordinate space as defined by its **origin** point.
Within the scope of `child`, point `p` has the coordinates as they are shown in the figure.
In the scope of `parent`, you have to add the offset of the `child` origin `O` to the coordinates of point `p`.
The resulting coordinates of `p` will then be:
\f{align*}{
p_{parent, 3d} &= \begin{bmatrix} 2.5 \\1.5 \\0.0 \end{bmatrix}
\f}

An additional offset of the `child` in `Z` direction is applied to the coordinate transform in the same manner. It just adds the offset to the resulting `Z` coordinate.

> The **origin** point of a *parent* is always `[0, 0, 0]` when dealing with the **parent scope**! When the *parent* is contained within another parent, its scope becomes the *child scope*. &rarr; This paradigm shift is probably the most confusing and hard to imagine part of this library, but it is important to keep the library flexible!

### Transform from 3D to 3D {#euler_angles}
So far no rotation was involved. The coordinate system had always the same orientation but where just translated within the 3D space.
That is not enough to represent arbitrary geometry. You need a mechanism to orient surfaces which should be extruded in a different direction than the global `Z` axis. For this reason, the idea of the **normal direction** is introduced.

The normal direction basically defines the direction where the local `Z` axis should point to within a parent scope.
The direction is defined using the [Direction_3](@ref geom2.Direction_3) class of **CGAL** and is a vector in 3D space. Although, this class is technically not concerned about the length of the vector, it is a good idea to make sure, that the resulting length of the normal direction is equal to **1.0**.

The definition of this normal direction is not enough to unambiguously define the three Euler angles which are needed for the coordinate transform. The normal direction can only define **two** of the three angles. As a consequence, the third Euler angle \f$\gamma\f$ has to be set manually using the `rotation_z` property of the geom2::Entity3D class.
This angle applies a rotation around the local `Z` axis whenever the coordinates of the geometry are transformed to another coordinate system.

@note The handling of the third Euler angle can still be subjected to change.

The usage of the Euler angles can lead to unintuitive results, but that is unfortunately the nature of those angles. Here are some results of Euler angles with different normal directions:
|Normal Direction | Euler Angle \f$\alpha\f$ | Euler Angle \f$\beta\f$ | Euler Angle \f$\gamma\f$ |
| --- | :---: | :---: | :---: |
|`[0, 0, 1]` | 0° | 0° | 0° |
|`[0, 1, 1]` | -45° | 0° | 0° |
|`[1, 0, 1]` | 45° | 0° | 90° |

Understanding the relationship of the normal direction and the Euler angles of this table is key to understand the 3D coordinate transformation with rotation.
**Most importantly,** the last case, where a "simple" rotation around the `Y` axis leads to *two* non-zero angles, especially the third angle \f$\gamma\f$ is not zero any more as opposed to the initial assumption! That is again due to the fact how Euler angles work...

In the end, you do not need to understand this in great detail to use the library, just be aware, that certain normal directions can lead to a final orientation which was not the one you expected. In such cases, you can use the `rotation_z` parameter to adjust.

## Class Diagram
Here is an overview how the library is structured:

![](class_diagram.png){html: width=1000}

@note This is still work in progress and can change!
