/*  Copyright (C) 2010-2023 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef BLACKBOXTEST_INCLUDE_BLACKBOXTEST_BLACKBOXTEST_H_
#define BLACKBOXTEST_INCLUDE_BLACKBOXTEST_BLACKBOXTEST_H_

#include <aixml/node.h>
#include <gtest/gtest.h>
#include <cstdio>
#include <unordered_map>
#include <string>
#include <utility>
#include <filesystem>
#include <variant>

using parameter_value = std::variant<std::string_view, double>; /**< Type of assertion value. Can be string_view or double */

/* Blackbox configuration */
constexpr const char* blackboxTestCasesXmlPath = "test/blackboxTestCases.xml"; /**< Path to the blackbox test cases xml file */
constexpr bool human_readable_output = true; /**< Flag to set the output to human readable format */

// https://google.github.io/googletest/advanced.html
class BlackboxTest : public ::testing::Test {
 protected:
    std::string tool_name{""}; /**< Name of the tool which is tested*/
    node* test_acXML; /**< Pointer to aircraft object the testing calculations are applied to */
    node& configXML; /**< Config XML node */
    static node* blackboxTestData; /**< Pointer to xml node containing blackbox testing data*/
    std::string currentTestSuiteName; /**< Name of currently running test suite */
    std::string currentTestCaseName; /**< Name of the currently executed testcase */
    static int8_t currentAssertion; /**< Variable with enumerated values for variable comparisons, e.g. GE (greater equal), LE, NEAR, etc. */
    std::pair <std::string, std::string> aircraftXmlParameterPair; /**< A pair of a parameter which will be set at a certain path in the aircraftXml */
    std::string preTestValue; /**< Original value of an entry within the aircraft.xml, which is temporarily changed to a test input value */

    struct TestCaseData; /**< Forward declaration of test case data */
    static std::unordered_map<std::string, TestCaseData>& getTestCaseMap(); /**< A map of test cases. The keys are the test case names */
    static std::unordered_map<std::string, std::string>& getTestParamPathsMap(); /**< Map of paths of test case parameters. The keys are the test case names */

    /** \brief Function adds a test case structure with a name and a node pointer to the test case map
     * \param name Name of the respective test case
     * \param testCaseNode Node pointer of the respective test case data
     * \return void
     */
    static void addCaseToMaps(std::string name, node* testCaseNode);


    /** \brief Function adds path for a respective test to the test case data maps
     * \param name Name of the respective test case
     * \param testNode Node pointer to the test case in the blackboxTestCases.xml file
     * \return void
     */
    static void addTestPathToMaps(std::string name, node* testNode);

    /** \brief Function to set up gtest and relevant testing data before every mockMain() execution
     */
    void SetupTest();
    /** \brief Function to clean up gtest and relevant testing data after every mockMain() execution
     */
    void TeardownTest();
    /** \brief Function applies every assertion given in blackbox testing xml file
        \param currentTestCaseNode Pointer to node object containing information on current test case
     */
    void applyAssertions(const node& criticalParamsNode, node* currentTestCaseNode, const std::string &testParamPath);
    virtual ~BlackboxTest() {} /**< Virtual destructor of BlackboxTest class*/

 public:
    explicit BlackboxTest(const std::string& tool_name); /**< Constructor of BlackboxTest class */
    /** \brief Function initializes gtest and runs all defined blackbox tests
     *  \param argc Number of command line arguments. Passed directly from main(...)
     *  \param argv[] Char array containing command line arguments. Passed directly from main(...)
     *  \param tool_name Name of the tool which is tested
     */
    static int runAllTests(int argc, char *argv[], std::string tool_name);
    /** \brief Function overrides gtest functionality to make standard testing procedures available to dynamic test registering
     *  \details Function is called by dynamically registered test to run test functionality.
     */
    void TestBody() override;
    /** \brief Function runs standard test functionality in 5 phases on current Test node
     */
    void standardTest();
    /** \brief Function runs standard death-test functionality in 5 phases on current Test node
     */
    void standardDeathTest();
    /** \brief Register the aircraft xml node
     *  \param aircraftNode Node containing the aircraft xml
     */
    void registerAircraft(node* aircraftNode);

 private:
     /** \brief Function to dynamically register all tests defined in .xml file with gtest
      *  \param tool_name Name of the tool which is tested
     */
    static void registerTests(const std::string& tool_name);
    /** \brief Function emulates relevant module behavior to simulate a full module execution
     */
    void executeModule();
    /** \brief Function parses assertion name into switch-accessible enum-based integer
     *  \param readValue Name of assertion to be converted
     *  \return theAssertion Integer value encoding the assertion, based on supportedAssertions enum
     */
    static int8_t getAssertion(const std::string& readValue);
    /** \brief Function checks whether a string is a number or not
     * \param str String to be evaluated
     * \return bool Boolean value. True if string is a number. False otherwise.
     */
    bool isNumber(const std::string& str);

    /** \brief Function returns a string with an error message
     * \param assertType Type of the assertion which is thrown
     * \param testParamPath Path, where the test parameter lies
     * \param testParamValue Value of the parameter which was altered
     * \param criticalParamPath Path to output parameter which is tested
     * \param criticalParamValue Value of the output parameter which was tested
     * \param assertion Value Assertion code
     * \param acc Accuracy of the tested value
     * \return Error message
     */
    std::string generateErrorMessage(int8_t assertType, const std::string &testParamPath,
                                                                   const std::string &testParamValue, const std::string &criticalParamPath,
                                                                   parameter_value criticalParamValue, const std::string &assertionValue, double acc = 0);
    /** \brief Function returns a string with an error message
     * \param inducedDeathTest Boolean value indicating whether the respective test is a normal test or an induced death test
     * \param testParamPath Path, where the test parameter lies
     * \param testParamValue Value of the parameter which was altered
     * \return Error message
     */
    std::string generateErrorMessage(bool inducedDeathTest, const std::string &testParamPath, const std::string &testParamValue);

    std::string getHumanReadableErrorMessage(int8_t assertType, const std::string &testParamPath,
                                                                   const std::string &testParamValue, const std::string &criticalParamPath, parameter_value criticalParamValue,
                                                                   const std::string &assertionValue, double acc);
    std::string getHumanReadableErrorMessage(bool inducedDeathTest, const std::string &testParamPath, const std::string &testParamValue);
    std::string getMachineReadableErrorMessage(void);
    /** \brief Class to control blackbox testing environment behavior
     */
    class Environment : public ::testing::Environment {
     public:
            ~Environment() override {} /** Virtual destructor of Environment class*/
            /** \brief Function called every time all blackbox tests are executed. Used to read testing data from .xml file
             */
            void TearDown() override;
    };

    /** \brief Class to generate modified human readable test logging in command line
        \detailed In 10/2022 it is impossible to inherit from testing::internal::PrettyUnitTestResultPrinter to only overwrite the functions one wants to change.
                  Therefore a wrapper class is written which calls the functions of a given EventListener in all functions where a standard functionality is desired.
     */
    class BetterPrinter : public testing::TestEventListener {
     private:
            TestEventListener* other;
     public:
            explicit BetterPrinter(TestEventListener * other);
            virtual ~BetterPrinter() {}
            void OnTestProgramStart(const testing::UnitTest& unit_test);
            void OnTestIterationStart(const testing::UnitTest& unit_test, int iteration);
            void OnEnvironmentsSetUpStart(const testing::UnitTest& unit_test);
            void OnEnvironmentsSetUpEnd(const testing::UnitTest& unit_test);
            void OnTestSuiteStart(const testing::TestSuite& test_suite);
            void OnTestStart(const testing::TestInfo& test_info);
            void OnTestPartResult(const testing::TestPartResult& test_part_result);
            void OnTestEnd(const testing::TestInfo& test_info);
            void OnTestSuiteEnd(const testing::TestSuite& test_suite);
            void OnEnvironmentsTearDownStart(const testing::UnitTest& unit_test);
            void OnEnvironmentsTearDownEnd(const testing::UnitTest& unit_test);
            void OnTestIterationEnd(const testing::UnitTest& unit_test, int iteration);
            void OnTestProgramEnd(const testing::UnitTest& unit_test);
    };

    /** List of available assertions
     */
    enum supportedAssertions{
        EQ, /**< EXPECT_EQ */
        STREQ, /**< EXPECT_STREQ */
        LE, /**< EXPECT_LE */
        GE, /**< EXPECT_GE */
        LT, /**< ECPECT_LT */
        GT, /**< EXPECT_GT */
        NEAR, /**< EXPECT_NEAR */
        DEATH /**< EXPECT_EXIT/EXPECT_DEATH */
        };
};

#endif // BLACKBOXTEST_INCLUDE_BLACKBOXTEST_BLACKBOXTEST_H_
