/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "engine/engine.h"
#include <aixml/node.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>
#include <cmath>
#include "SI_units.h"

/* === Constructors === */
Engine::Engine(const std::filesystem::path &engine_directory)
    :
    engine_data_{
          aixml::openDocument(
              engine_directory / (engine_directory.filename().string() + ".xml"))},
    op_{this->engine_data_.design_point()},
    enginetype_{"turbofan"},    // TODO(Oli): needs to be read from engine.xml (V3) in the future
    scale_factor_{1.0},
    number_of_engines_{1},
    N_correction_{1.0},
    W_correction_{1.0},
    T_correction_{1.0},
    N_shaft_offtake_bleed_correction_{1.0},
    W_fuel_shaft_offtake_bleed_correction_{1.0},
    temperature_shaft_offtake_bleed_correction_{1.0},
    is_derate_increased_{false},
    N_upper_limit_{NAN},
    N_lower_limit_{NAN},
    thrust_correction_bleed_{1.0},
    // Constants for the calculation process
    N1_step_{0.005},

    // Engine Limits
    WF_to_P3_min_{engine_data_.fuelflow_to_p3_ratio()},
    WF_to_P3_max_TO_{engine_data_.fuelflow_to_P3max_ratio_at_MTO()},
    WF_to_P3_max_MCT_{engine_data_.fuelflow_to_P3max_at_MCT()},
    max_temperature_TO_{engine_data_.max_temperature_MTO()},
    max_temperature_continuous_thrust_{engine_data_.max_temperature_MCT()},
    N2_max_{engine_data_.max_N2()},
    max_power_TO_{0.0},    // TODO(Oli): is only non zero value, if turboprop
    max_power_continuous_thrust_{0.0},    // TODO(Oli): is only non zero value, if turboprop
    N_fraction_climb_rating_{engine_data_.fraction_N_to_climbrating()},
    N_fraction_cruise_rating_{engine_data_.fraction_N_to_cruiserating()},
    max_relative_bleed{engine_data_.max_relative_bleed()},
    max_shaft_power_extraction_{engine_data_.max_shaft_power_extraction()},
    unscaled_SLST_{engine_data_.unscaled_SLST()},
    dry_mass_{engine_data_.dry_mass()},

    // Engine deck names
    thrustdeck{engine_data_.name() + "_FN"},            //thrust
    fuelflow{engine_data_.name()+"_WF"},                //fuel flow
    core_mass_flow{engine_data_.name()+"_W25"},         //core mass flow
    overall_mass_flow{engine_data_.name()+"_St2_W"},         //Overall mass flow
    sNOx{engine_data_.name()+"_sNOx"},                  //severity NOX parameter
    fraction_fuelflow_to_p3{engine_data_.name()+"_WFqP3"}, // Ratio of fuel flow to P3
    temperature_limit_deck{engine_data_.name()+"_St5_T"},  //temperature limit
    //temperature_limit_deck{engine_data_.get_get_temperature_limit_deck_name()},
    N_limit_deck{engine_data_.name()+"_XN_HPC"}         // rel values of N2
{
}

Engine::Engine(const std::filesystem::path &engine_directory, const double scale_factor)
: Engine{engine_directory}
{
    this->scale_factor_ = scale_factor;
}

// === Getters ===
auto Engine::get_unscaled_engine() const -> const EngineData &
{
    return this->engine_data_;
}

auto Engine::get_scale_factor() const -> double
{
    return this->scale_factor_;
}

auto Engine::get_engine_dimensions() const -> Dimensions
{
    // The dimensions as defined in the XML file
    auto dimensions_unscaled = this->engine_data_.dimensions();
    // Scale the dimensions using the thrust scaling.
    // => The cross-section parameters are
    return {
        dimensions_unscaled.height * std::pow(this->scale_factor_, 0.5),
        dimensions_unscaled.width * std::pow(this->scale_factor_, 0.5),
        dimensions_unscaled.length * std::pow(this->scale_factor_, 0.4),
        dimensions_unscaled.diameter * std::pow(this->scale_factor_, 0.5)
    };
}

auto Engine::get_operating_point() -> OperatingPoint
{
    this->op_.N *= this->N_correction_ * this->N_shaft_offtake_bleed_correction_;
    return this->op_;
}

auto Engine::get_same_engine_type_quantity() -> int
{
    return this->number_of_engines_;
}

// === Setters ===
void Engine::set_operating_point(const OperatingPoint &op)
{
    this->op_ = op;
}

void Engine::set_same_engine_type_quantity(int16_t same_engine_type_quantity)
{
    if (same_engine_type_quantity < 0) {
        std::cout << "WARNING: ";
        std::cout << "'" << same_engine_type_quantity << "' number of engines not possible. ";
        std::cout << "Assuming '+" << same_engine_type_quantity * (-1) << "' number of engines instead." << std::endl;
        same_engine_type_quantity = same_engine_type_quantity * (-1);
    }
    this->number_of_engines_ = same_engine_type_quantity;
}

void Engine::set_correction_factors(const OperatingPoint &op, const atmosphere& atm)
{
    // Calculate correction values for N1, mass flow, and Temperature
    double temp = atm.getTemperature(this->op_.altitude);
    double tempISA = atm.getTemperatureISA(this->op_.altitude);
    double pressure = atm.getPressure(this->op_.altitude);
    double pressureISA = atm.getPressureISA(this->op_.altitude);

    N_correction_ = sqrt(temp / tempISA);
    W_correction_ = sqrt(tempISA / temp) * (pressure / pressureISA);
    T_correction_ = temp / tempISA;
}

void Engine::calculate_N1_with_penalties(const double& flight_level,
                                         const double& mach_number,
                                         const atmosphere& atm,
                                         const double& derate,
                                         const std::string& thrust_rating,
                                         double bleed_air_offtake,
                                         double shaft_power_offtake)
{
    // Set environmental values
    this->op_.altitude = flight_level;
    this->op_.Mach = mach_number;
    shaft_power_offtake /= this->number_of_engines_;
    bleed_air_offtake /= this->number_of_engines_;
    this-> set_correction_factors(op_, atm);
    // Calculate operating limits (Idle N1) w.r.t. operating point
    this->set_lower_N1_limit();
    double residualNmom(1.);
    double itCounter(0.);
    while (residualNmom > ACCURACY_HIGH) {    // Converge N_mom to 1e-10
        double currentNmom(this->op_.N);
        // Calculate operating limits (Max N1) w.r.t. operating point, temperature and overall compression ratio limits
        // --> Initially without correction of bleed air and shaft power extractions
        this->set_upper_N1_limit(thrust_rating);
        // Set current low pressure shaft speed (N_mom) w.r.t Lower-/Upper-Limit and thrust rating
        this->set_N1_to_rating(thrust_rating, derate);
        // Calculate corrections for N1 due to bleed & shaft power-offtakes
        this->calculate_offtake_correction(atm, bleed_air_offtake, shaft_power_offtake);
        // Calculate operating limits (Max N1) w.r.t. operating point, temperature and overall compression ratio limits
        // --> Now WITH correction of bleed air and shaft power extractions
        this->set_upper_N1_limit(thrust_rating);
        // Set current low pressure shaft speed (N_mom) w.r.t Lower-/Upper-Limit and thrust rating
        this->set_N1_to_rating(thrust_rating, derate);
        residualNmom = (fabs(currentNmom - this->op_.N) / currentNmom);
        if (itCounter > 500) {
            if (residualNmom > ACCURACY_MEDIUM) {
                std::stringstream errMsg;
                errMsg << "Could not find low pressure shaft speed N1 for current rating: ";
                errMsg << thrust_rating << std::endl;
                errMsg << "N1(current): " << this->op_.N << "; ";
                errMsg << "N1(last): " << currentNmom << "; ";
                errMsg << "N1(upperLimit): " << this->N_upper_limit_;
                throwError(__FILE__, __func__, __LINE__, errMsg.str());
            } else {
                break;
            }
        }
        ++itCounter;
    }
    if (is_derate_increased_) {
        is_derate_increased_ = false;
        throw 1;
    }
}

void Engine::check_offtake_boundaries(const OperatingPoint &op, double bleedAir, double shaftPower)
{
    double coreMassFlow{engine_data_.get_deck_value(core_mass_flow, this->op_) * scale_factor_ / this->W_correction_};
    double relBleed{1.06 * bleedAir / coreMassFlow};
    if (relBleed > this->get_max_bleed_offtake_at_current_operating_point()) {
        std::stringstream errMsg;
        errMsg << "Relative bleed air extraction (" << relBleed << ") too large. ";
        errMsg << "Max. allowed extraction is " << this->max_relative_bleed << "." << std::endl;
        errMsg << "Altitude: " << this->op_.altitude << " m; ";
        errMsg << "Mach: " << this->op_.Mach << "; ";
        errMsg << "N1: " << this->op_.N;
        throwError(__FILE__, __func__, __LINE__, errMsg.str());
    }
    if (shaftPower > get_max_shaft_power_offtake_engine()) {
        throwError(__FILE__, __func__, __LINE__, "Shaft power extraction too large!");
    }
}

auto Engine::converge_value(const OperatingPoint &op,
                            double goal_value,
                            double iteration_step,
                            std::string deckvalue) -> double
{
    // Initialize variables for interpolation
    const int MAX_ITERATIONS{50};   // Assuming a max iteration count of 50 as value should be reached
    double delta, x1, x2, y1, y2, y, x_new;
    bool value_should_increase{false};
    int iterationcount{0};
    double step = iteration_step;

    // Helper lambda to fetch thrust or deck value
    auto get_thrust_or_deck_value = [&]() -> double {
        return (deckvalue == "_FN") ? get_thrust() :
            this->engine_data_.get_deck_value(engine_data_.name() + deckvalue, this->op_);
    };

    // Only go through loop if the delta is there
    double value_at_op = get_thrust_or_deck_value();
    delta = std::abs((goal_value-value_at_op)/value_at_op);
    // Loop to get to a convergence
    while (delta > ACCURACY_MEDIUM) {
        x1 = get_thrust_or_deck_value();
        y1 = op_.N;
        value_should_increase = (x1 < goal_value);
        // Initialize variable for iteration to be + or - step
        y2 = value_should_increase ? (op.N + step) : (op.N - step);
        // Check that the operating point is within the limits. Operating points must allow goal value
        op_.N = std::clamp(y2, this->N_lower_limit_, this->N_upper_limit_);
        // get linear interpolation to be on the point where we expect the value
        x2 = get_thrust_or_deck_value();
        // Interpolate
        y = y1 + ((y2 - y1) / (x2 - x1)) * (goal_value - x1);
        this->op_.N = y;
        // check if it is close enough otherwise set the limit for the new interpolation and do it again
        op_.N = std::clamp(y, this->N_lower_limit_, this->N_upper_limit_);
        x_new = get_thrust_or_deck_value();
        // Calculate delta
        delta = std::abs((x_new-goal_value)/goal_value);
        // Set new boundaries
        step /= 2.0;
        // Handel iteration boundaries
        if (++iterationcount > MAX_ITERATIONS) {
            std::cout << "Cannot converge to N1" << std::endl;
            throw 1;
        }
        if (y >= N_upper_limit_) {
            std::cout << "Reached upper limit: " << N_upper_limit_ << std::endl;
            op_.N = N_upper_limit_;
            break;
        }
        if (y <= N_lower_limit_) {
            std::cout << "Reached lower limit: " << N_lower_limit_ << std::endl;
            op_.N = N_lower_limit_;
            break;
        }
    }
    return op_.N;
}

void Engine::calculate_N1_with_thrustlimit(const double& flight_level,
                                           const double& mach_number,
                                           const atmosphere& atm,
                                           const double& derate,
                                           const std::string& thrust_rating,
                                           double bleed_air_offtake,
                                           double shaft_power_offtake,
                                           double thrust_limit) {
    // Set engine rating
    this->calculate_N1_with_penalties(flight_level,
                                      mach_number,
                                      atm,
                                      derate,
                                      thrust_rating,
                                      bleed_air_offtake,
                                      shaft_power_offtake);
    // Set thrust limit
    thrust_limit /= this->number_of_engines_;
    this->calculate_offtake_correction(atm, bleed_air_offtake, shaft_power_offtake);

    // Engine deck names
    double thrust_at_currentN1_tmp = get_thrust();
    while (thrust_limit > thrust_at_currentN1_tmp)
    {
        this->op_.N += N1_step_;
        if (this->op_.N > this->N_upper_limit_) {
            this->op_.N = this->N_upper_limit_;
            // If upper limit is reached it makes no sense to continue while loop here
            // Throw error so mission can deal with it. It knows what to do with these numbers
            double thrust_at_max_N1_tmp = get_thrust();
            if (thrust_limit > thrust_at_max_N1_tmp) {
                // std::cout << "Engine unable to deliver the required thrust. Max thrust:" << thrust_at_max_N1_tmp << "N" << std::endl;
                throw 2;    // Mission knows what to do with that
                }
            }
        thrust_at_currentN1_tmp = get_thrust();
    }

    while (thrust_limit < thrust_at_currentN1_tmp)
    {
        this->op_.N -= N1_step_;
        if (this->op_.N < this->N_lower_limit_) {
            this->op_.N = this->N_lower_limit_;
            // If lower limit is reached it makes no sense to continue while loop here
            // Throw error so mission can deal with it. It knows what to do with these numbers
            double thrust_at_min_N1_tmp = get_thrust();    // get the thrust at the lowest possible shaft speed [N]
            if (thrust_limit < thrust_at_min_N1_tmp) {
                throw 3;    // Mission knows what to do with that
            }
        }
        thrust_at_currentN1_tmp = get_thrust();
    }
    op_.N = this-> converge_value(op_, thrust_limit, N1_step_, "_FN");

        // Calculate corrections for N_mom due to bleed & shaft power-offtakes
        shaft_power_offtake /= this->number_of_engines_;
        bleed_air_offtake /= this->number_of_engines_;
        this->calculate_offtake_correction(atm, bleed_air_offtake, shaft_power_offtake);

    if (is_derate_increased_) {
        is_derate_increased_ = false;
        throw 1;
    }
}

// === Access deck values ===
auto Engine::get_thrust_aircraft() -> double
{
    // Get the deck value and scale it directly with the thrust scale factor
    return SI::force(this->engine_data_.get_deck_value(thrustdeck, this->op_), "kN") *
        this->scale_factor_ *
        this->number_of_engines_ *
        thrust_correction_bleed_;
}

auto Engine::get_thrust() -> double
{
    // Get the deck value and scale it directly with the thrust scale factor
    return SI::force(this->engine_data_.get_deck_value(thrustdeck, this->op_), "kN") *
        this->scale_factor_; //  * thrust_correction_bleed_;
}

auto Engine::get_aircraft_fuelflow() -> double
{
    // Get the deck value and scale it directly with the thrust scale factor, since TSFC stays constant
    return this->engine_data_.get_deck_value(fuelflow, this->op_) *
      this->W_fuel_shaft_offtake_bleed_correction_ *
      this->scale_factor_;
}

auto Engine::get_fuelflow() -> double
{
    // Get the deck value and scale it directly with the thrust scale factor, since TSFC stays constant
    return this->engine_data_.get_deck_value(fuelflow, this->op_) *
      this->W_fuel_shaft_offtake_bleed_correction_ *
      this->scale_factor_;
}

auto Engine::get_engine_EGT() -> double
{
    return this->engine_data_.get_deck_value("St5_temperature", this->op_) * T_correction_ * temperature_shaft_offtake_bleed_correction_;    // TODO(Tobi): check .csv name
}

auto Engine::get_engine_EPR() -> double
{
    return this->engine_data_.get_deck_value("EPR", this->op_);    // TODO(Tobi): check .csv name;
}

auto Engine::get_engine_OPR() -> double
{
    return this->engine_data_.get_deck_value("OPR", this->op_);    // TODO(Tobi): check .csv name;
}

auto Engine::get_engine_NOx_emission_index() -> double
{
    // Factor 1000 due to GasTurb deck being in g/kg. 
    // TODO(Tobi): Check .csv name
    return this->engine_data_.get_deck_value(sNOx, this->op_) * get_engine_severity_parameter() * 1000;
}

auto Engine::get_LTO_emission_index(const LTOPhases LTO_cycle, const EngineEmissions emission) -> double
{
    switch (emission)
    {
    case EngineEmissions::HC :
        return this->engine_data_.LTO_emission_HC(LTO_cycle);
    case EngineEmissions::CO :
        return this->engine_data_.LTO_emission_CO(LTO_cycle);
    case EngineEmissions::NOx :
        return this->engine_data_.LTO_emission_NOx(LTO_cycle);
    case EngineEmissions::SN :
        return this->engine_data_.LTO_emission_SN(LTO_cycle);
    default:
        return 0.0;
    }
}

auto Engine::get_LTO_fuelflow(const LTOPhases LTO_cycle) -> double
{
    // The LTO fuel flow is now scaled with the scale factor according to the scaling of the engine fuel flow
    return this->engine_data_.LTO_fuel_flow(LTO_cycle) *
    this->scale_factor_;
}

auto Engine::get_max_shaft_power_offtake_engine() -> double
{
    // the maximal allowed shaft power extraction scales linearly with the scale factor
    return this->max_shaft_power_extraction_ * scale_factor_;
}

auto Engine::get_max_bleed_offtake_at_current_operating_point() -> double
{
    // the maximal allowed bleed air offtake scales linearly with the thrust_scale_factor
    return this->max_relative_bleed * engine_data_.get_deck_value(core_mass_flow, this->op_) * scale_factor_;
}

auto Engine::get_dry_mass() -> double 
{
    // get the scaled dry mass of the engine according to Raymer p. 284 
    return this->dry_mass_ * pow(scale_factor_,1.1); 
}

auto Engine::get_physical_properties_stage(const StageProperties stage_property,
      const EngineStage engine_stage) -> double
      {
        // convert the input nums into two temporary strings to concatenate them later
        std::string tmp_string_stage_property{};
        std::string tmp_string_engine_stage{};

        // convert enums to strings.
        switch (stage_property)
        {
        case StageProperties::FuelToAirRatio: tmp_string_stage_property = "fa"; break;
        case StageProperties::MachNumber: tmp_string_stage_property = "Mn"; break;
        case StageProperties::TotalPressure: tmp_string_stage_property = "P"; break;
        case StageProperties::TotalTemperature: tmp_string_stage_property = "T"; break;
        case StageProperties::StaticTemperature: tmp_string_stage_property = "Ts"; break;
        case StageProperties::MassFlow: tmp_string_stage_property = "W"; break;
        default:
            std::cout << "Unknown engine stage property." << std::endl;
            return 0.0;
        }

        switch (engine_stage)
        {
        case EngineStage::St2 : tmp_string_engine_stage = "_St2_"; break;
        case EngineStage::St3 : tmp_string_engine_stage = "_St3_"; break;
        case EngineStage::St4 : tmp_string_engine_stage = "_St4_"; break;
        case EngineStage::St5 : tmp_string_engine_stage = "_St5_"; break;
        case EngineStage::St13: tmp_string_engine_stage = "_St13_"; break;
        case EngineStage::St22: tmp_string_engine_stage = "_St22_"; break;
        case EngineStage::St25: tmp_string_engine_stage = "_St25_"; break;
        case EngineStage::St45: tmp_string_engine_stage = "_St45_"; break;
        case EngineStage::St18: tmp_string_engine_stage = "_St18_"; break;
        case EngineStage::St8 : tmp_string_engine_stage = "_St8_"; break;
        default:
            std::cout << "Unknown engine stage." << std::endl;
            return 0.0;
        }

        // concatenate the tmp string tho match the .cvs name
        return engine_data_.get_deck_value(engine_data_.name() + tmp_string_engine_stage + tmp_string_stage_property,
                                           op_);
      }

// === PRIVATE FUNCTIONS ===
auto Engine::get_engine_severity_parameter() -> double
{
    return engine_data_.get_NOx_severity_parameter();
}

auto Engine::get_scaled_SLST() -> double
{
    return SI::force(this->unscaled_SLST_, "kN") * this->scale_factor_;
}

void Engine::set_lower_N1_limit() {
    // Estimation of ground/flight idle
    // FADEC sets minimum fuel flow to prevent flame-out (WF/P3)
    // Set lower N1 limit to minimum required fuel
    // (first converged N1 in fuel flow deck for current altitude and Mach number)
    this->N_lower_limit_ = op_.N = engine_data_.get_lower_operating_point(fuelflow).N;
    // N_lower_limit_ is increased from min value in deck that shows all valid operating points to first valid value
    while (engine_data_.get_deck_value(fuelflow, this->op_) < 0)
    {
        op_.N += N1_step_;
        this->N_lower_limit_ = op_.N;    // update the N_lower_limit_ variable
    }
    while (engine_data_.get_deck_value(thrustdeck, this->op_) < 0)
    {
        op_.N += N1_step_; 
        this->N_lower_limit_ = op_.N; // update the N_lower_limit_ variable
    }
    // Initialize check variables
    // Check flameout limit deck (WFqP3) if flameout danger is true
    // and ratio of fuelflow to combustion chamber inlet is below limit.
    // In this case N1 is increased
    while (this->engine_data_.get_deck_value(fraction_fuelflow_to_p3, this->op_) < this->WF_to_P3_min_) {
        op_.N += N1_step_;
        this->N_lower_limit_ = op_.N;    // update N_lower_limit_ variable
        if (this->N_lower_limit_ > this->N_upper_limit_) {
            throwError(__FILE__, __func__, __LINE__, "Idle in WFToP3-limit deck not found.");
        }
    }
}

void Engine::set_upper_N1_limit(const std::string &thrust_rating) {
    // Estimation of upper N1 value
    // FADEC sets maximum fuel flow to prevent over temperatures as a limit of temperatures, pressures, HP spool speed
    // Set upper N1 limit to maximum required fuel
    // (last converged N1 in fuel flow deck for current altitude and Mach number)
    this->N_upper_limit_ = op_.N = engine_data_.get_upper_operating_point(fuelflow).N;    // Reset upper N1 limit
    // N_upper_limit_ is decreased from max value in valid operating point deck to first valid value
    while (engine_data_.get_deck_value(fuelflow, this->op_) < 0)
    {
        this->op_.N -= N1_step_;
        this->N_upper_limit_ = this->op_.N;    // update the N_upper_limit_ variable
    }
    // Temperature limit
    this->set_upper_N1_limit_temperature(thrust_rating);
    // FuelFlow/Pressure limit
    this->set_upper_N1_limit_with_overall_fuelflow_pressure_ratio(thrust_rating);
    // HP Spool speed limit
    this->set_upper_N1_limit_high_pressure_spool_speed(thrust_rating);
    // Power limit
    if (this->enginetype_ == "prop") {
        this->set_upper_N1_limit_power(thrust_rating);
    }
}

void Engine::calculate_offtake_correction(const atmosphere& atm, const double& bleedAir, const double& shaftPower) {
    // The factors for correction were done according to the CF6 installation manual.
    // Range of values: Mach: 0-0.95; Relative corrected thirst: 0.2-1
    // create temporary operating point to feed into the function
    this->check_offtake_boundaries(op_, bleedAir, shaftPower);
    double coreMassFlow{engine_data_.get_deck_value(core_mass_flow, this->op_) * scale_factor_ / this->W_correction_}; 
    double overall_mass_flow_{engine_data_.get_deck_value(overall_mass_flow, this->op_) * scale_factor_ / this->W_correction_};
    double relBleed{1.06 * bleedAir / coreMassFlow};
    thrust_correction_bleed_ = 1 - (2 * ( bleedAir / overall_mass_flow_)); // Ray p. 488
    // Constants and Intermediate Variables
    const double tempAtAltitude = atm.getTemperature(this->op_.altitude);
    const double tempISA = atm.getTemperatureISA(0.0);
    const double pressureAtAltitude = atm.getPressure(this->op_.altitude);
    const double pressureISA = atm.getPressureISA(0.0);

    // Calculate correction factors
    // Statistical fit from student thesis.
    // Paul Pross 2009 "Entwicklung einer Methode zur Dimensionierung von Flugantrieben im Flugzeugvorentwurf"
    double referenceMaxShaftPower{((0.27 * get_scaled_SLST() / 1000) + 5.0) * 1000};
    double relCorrectedShaftPower{shaftPower /
                                  sqrt(tempAtAltitude / tempISA) /
                                  (pressureAtAltitude / pressureISA) /
                                  referenceMaxShaftPower};
    double relCorrectedThrust{std::min(1.0, ((get_thrust()) /
                                            (pressureAtAltitude / pressureISA)) / get_scaled_SLST())};
    
    // Calculate the correction using the adapted bspline
    N_shaft_offtake_bleed_correction_ = 1.0 -
      correctionNBleed(this->op_.Mach, relCorrectedThrust) * relBleed -
      correctionNPshaft(this->op_.Mach, relCorrectedThrust) * relCorrectedShaftPower;
    N_shaft_offtake_bleed_correction_ = std::max(1.0, N_shaft_offtake_bleed_correction_); // TODO(Tobi): Do we still need htis comment? // Fix to not let it crash when only shaft power is extruded - minimum value possible was 0.9985

    W_fuel_shaft_offtake_bleed_correction_ = 1.0 -
      correctionFuelBleed(this->op_.Mach, relCorrectedThrust) * relBleed -
      correctionFuelPshaft(this->op_.Mach, relCorrectedThrust) * relCorrectedShaftPower;

    temperature_shaft_offtake_bleed_correction_ = 1.0 +
      correctionTemperatureBleed(this->op_.Mach, relCorrectedThrust) * relBleed +
      correctionTemperaturePshaft(this->op_.Mach, relCorrectedThrust) * relCorrectedShaftPower;
}

void Engine::set_N1_to_rating(const std::string& thrustRating, const double& derate) {
    if (derate > 1.0 || derate < 0.1) {
        std::stringstream errorMessage;
        errorMessage << "Invalid range!" << std::endl;
        errorMessage << "Derate factor ";
        errorMessage << derate << " (0.1 < X < 1.0 is allowed)" << std::endl;
        errorMessage << "Abort program!" << std::endl;
        throwError(__FILE__, __func__, __LINE__, errorMessage.str());
    }
    if (thrustRating == "takeoff" || thrustRating == "maximum_continuous") {
        this->op_.N = std::max(this->N_upper_limit_ * derate, this->N_lower_limit_);
        // std::stringstream errMsg;    // TODO(Oli): this should be a INFO
        // errMsg << "N_mom (TO/MaxCont): " << this->op_.N << std::endl;
        // throwError(__FILE__, __func__, __LINE__, errMsg.str());
        if (this->op_.N > this->N_upper_limit_ * derate) {
            this->is_derate_increased_ = true;
        }
    } else if (thrustRating == "climb") {
        this->op_.N = std::max(this->N_upper_limit_ * this->N_fraction_climb_rating_ * derate,
                               this->N_lower_limit_);
        // std::stringstream errMsg;    // TODO(Oli): this should be a INFO
        // errMsg << "N_mom (Climb): " << this->op_.N << std::endl;
        // throwError(__FILE__, __func__, __LINE__, errMsg.str());
        if (this->op_.N > this->N_upper_limit_ * this->N_fraction_climb_rating_ * derate) {
            this->is_derate_increased_ = true;
        }
    } else if (thrustRating == "cruise") {
        this->op_.N = std::max(this->N_upper_limit_ * this->N_fraction_cruise_rating_ * derate,
                               this->N_lower_limit_);
        // std::stringstream errMsg;    // TODO(Oli): this should be a INFO
        // errMsg << "N_mom (Cruise): " << this->op_.N << std::endl;
        // throwError(__FILE__, __func__, __LINE__, errMsg.str());
        if (this->op_.N > this->N_upper_limit_ * this->N_fraction_cruise_rating_ * derate) {
            this->is_derate_increased_ = true;
        }
    } else if (thrustRating == "idle") {
        this->op_.N = this->N_lower_limit_;
        // std::stringstream errMsg;    // TODO(Oli): this should be a INFO
        // errMsg << "N_mom (Idle): " << this->op_.N << std::endl;
        // throwError(__FILE__, __func__, __LINE__, errMsg.str());
        if (derate < 1.0) {
            this->is_derate_increased_ = true;
        }
    } else {
        std::stringstream errorMessage{};
        errorMessage << "Rating: " << thrustRating << " not known! Abort program!" << std::endl;
        throwError(__FILE__, __func__, __LINE__, errorMessage.str());
    }
}

void Engine::set_upper_N1_limit_temperature(const std::string &thrust_rating) {
    // Temperature limit for corrected temperature! flat-rated
    double TempLimit = (thrust_rating == "takeoff")
                   ? this->max_temperature_TO_
                   : this->max_temperature_continuous_thrust_;
    this->op_.N = this->N_upper_limit_;
    // N_upper_limit_ is decreased from max value in fuel deck to first valid value
    while (engine_data_.get_deck_value(temperature_limit_deck, this->op_) *
           this->T_correction_ *
           this->temperature_shaft_offtake_bleed_correction_ > TempLimit) {
        this->op_.N -= N1_step_;
        this->N_upper_limit_ = this->op_.N;    // update the N_upper_limit_ variable
        if (this->N_upper_limit_ < N_lower_limit_) {
            std::stringstream errorMessage{};
            errorMessage << "Engine N1 speed below idle speed by temp-limit! ";
            errorMessage << "Impossible operating point." << std::endl;
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
        }
    }
}

void Engine::set_upper_N1_limit_with_overall_fuelflow_pressure_ratio(const std::string &thrust_rating) {
    // FuelFlow/Pressure ratio limit
    double maximumPressureRatio = (thrust_rating == "takeoff")
                              ? this->WF_to_P3_max_TO_
                              : this->WF_to_P3_max_MCT_;
    this->op_.N = this->N_upper_limit_;
    // N_upper_limit_ is decreased from max value in fuel deck to first valid value
    while (engine_data_.get_deck_value(fraction_fuelflow_to_p3, op_) > maximumPressureRatio) {
        op_.N -= N1_step_;
        this->N_upper_limit_ = this->op_.N;    // update N_upper_limit_ variable
        if (this->N_upper_limit_ < this->N_lower_limit_) {
            std::stringstream errorMessage{};
            errorMessage << "Engine N1 speed below idle speed by compression-limit! ";
            errorMessage << "Impossible operating point." << std::endl;
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
        }
    }
}

void Engine::set_upper_N1_limit_high_pressure_spool_speed(const std::string &thrust_rating) {
    this->op_.N = this->N_upper_limit_;
    // N_upper_limit_ is decreased from max value in fuel deck to first valid value
    while (engine_data_.get_deck_value(N_limit_deck, this->op_) > this->N2_max_) {
        this->op_.N -= N1_step_;
        this->N_upper_limit_ = this->op_.N;    // Update N_upper_limit_ variable
        if (this->N_lower_limit_ > this->N_upper_limit_) {
            std::stringstream errorMessage{};
            errorMessage << "NupperLimit is below NlowerLimit. ";
            errorMessage << "No valid operation point." << std::endl;
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
        }
    }
}

void Engine::set_upper_N1_limit_power(const std::string &thrust_rating) {
    // Set power limit
    double powerLimit = (thrust_rating == "takeoff")
                    ? this->max_power_TO_
                    : this->max_power_continuous_thrust_;
    // create temporary operating point to feed into the function
    op_.N = this->N_upper_limit_;
    // N_upper_limit_ is decreased from max value in fuel deck to first valid value
    while (engine_data_.get_deck_value(fuelflow, op_) *
           this->T_correction_ *
           this->temperature_shaft_offtake_bleed_correction_ > powerLimit) {
        op_.N -= N1_step_;
        this->N_upper_limit_ = op_.N;    // update N_upper_limit_ variable
        if (this->N_upper_limit_ < N_lower_limit_) {
            std::stringstream errorMessage{};
            errorMessage << "Engine N1 speed below idle speed by power-limit! ";
            errorMessage << "Impossible operating point." << std::endl;
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
        }
    }
}

double Engine::calculate_polynom(const double coeffs[], const double x, const double y) {
    // Calculate the polynomial value using a bi-quadratic symmetric approach
    const double correction_factor = 
                         coeffs[0] +
                         coeffs[1] * x +
                         coeffs[2] * y +
                         coeffs[3] * x * x +
                         coeffs[4] * x * y +
                         coeffs[5] * y * y +
                         coeffs[6] * x * x * x +
                         coeffs[7] * x * x * y +
                         coeffs[8] * x * y * y +
                         coeffs[9] * y * y * y +
                         coeffs[10] * x * x * x * x +
                         coeffs[11] * x * x * x * y +
                         coeffs[12] * x * x * y * y +
                         coeffs[13] * x * y * y * y +
                         coeffs[14] * y * y * y * y;
         return correction_factor;
}

// Correction for shaft power offtake
double Engine::correctionNPshaft(const double& Mach, const double& relCorrectedThrust) {
    // Coefficients for the polynomial model based on a regression 
    double coeffs[] = {
        -8.31692668889181e-05,	0.000383140099801511,	0.00227319600772912,	-0.000176657994605840,	-0.00471618256285559,
        -0.00138076435912756,	3.48611110081685e-09,	0.00220819792267324,	0.00267411558816458,	-2.24192601196653e-09,
        -1.54536538694497e-09,	-1.81953666270857e-09,	-0.00107973284690613,	9.96622650499607e-11,	1.11182396189105e-09
    };

    return this->calculate_polynom(coeffs, relCorrectedThrust, Mach);
}

// Fuel correction
double Engine::correctionFuelPshaft(const double& Mach, const double& relCorrectedThrust) {
   // Coefficients for the polynomial model based on a regression 
    double coeffs[] = {
        0.0309718362077493,	-0.0599702637683423,	-0.0370829165209619,	0.0300515165960587,	0.0603807421071228,
        0.0144666441444089,	-2.11134251209117e-09,	-0.0266894766231007,	-0.0127903940847991,	2.15927023829018e-09,
        2.13274413987745e-09, 5.48382591709232e-09,	0.00107304658738337,	1.24437717675390e-09,	-1.04034385729637e-09
    };

    return this->calculate_polynom(coeffs, relCorrectedThrust, Mach);
}

// Temperature Correction
double Engine::correctionTemperaturePshaft(const double& Mach, const double& relCorrectedThrust) {
    // Coefficients for the polynomial model based on a regression 
    double coeffs[] = {
        0.0309718362077493,	-0.0599702637683423, -0.0370829165209619,	0.0300515165960587,	0.0603807421071228,
   	    0.0144666441444089,	-2.11134251209117e-09,	-0.0266894766231007,	-0.0127903940847991,	2.15927023829018e-09,
        2.13274413987745e-09,	-5.48382591709232e-09,	0.00107304658738337,	1.24437717675390e-09, -1.04034385729637e-09
    };

    return this->calculate_polynom(coeffs, relCorrectedThrust, Mach);
}

// Correction for spool speed due to bleed air
double Engine::correctionNBleed(const double& Mach, const double& relCorrectedThrust) {
    // Coefficients for the polynomial model based on a regression 
    double coeffs[] = {
        -0.0243533170897740,	0.0216467267229328,	    -0.141694321704478,	-0.0713989405375088, 0.425637647707458,
        0.0745660202730429,	    5.13941893190816e-09,   -0.617641501133532,	-0.401082822726307,	-3.05944936645471e-09,
        -2.32232190108554e-09,	-3.74633322233831e-10,	0.626050000913035,	1.06626580328299e-09,	1.65647059080228e-09
    };

    return this->calculate_polynom(coeffs, relCorrectedThrust, Mach);
}

// Correction for fuel flow
double Engine::correctionFuelBleed(const double& Mach, const double& relCorrectedThrust) {
    // Coefficients for the polynomial model based on a regression 
    double coeffs[] = {
       -2.06366228619814, 0.769847774981191, -0.357598736559579, -0.368851502212470, 1.62403220433678,
       -0.248890043582220, -9.38687038611999e-09, -1.45412806188745, -0.354588867666351, -4.50946724478958e-09,
       4.74563104811297e-09, 6.12516308073055e-11, 0.579120839303474, 1.11789642943655e-09, 1.89344285483325e-09
    };

    return this->calculate_polynom(coeffs, relCorrectedThrust, Mach);
}

// Temperature correction
double Engine::correctionTemperatureBleed(const double& Mach, const double& relCorrectedThrust) {
    // Coefficients for the polynomial model based on a regression
    double coeffs[] = {
        1.39153085298236, -0.502857981817937, 0.0988635530151397, 0.445617614551110, -0.445485483721538,
        -0.187193041316873, 1.74339556331964e-09, 0.673617823006167, 0.586640295078888, -8.95656368284025e-10,
        -6.78274442809335e-10, -2.64486207224415e-09, -0.676418234172954, -2.43471913547178e-09, 1.21904183283406e-09 
    };

    return this->calculate_polynom(coeffs, relCorrectedThrust, Mach);
}