/*  Copyright (C) 2010-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    Copyright (C) 2024 Chair of Aircraft Design, Technical University Munich
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * @file csv.cpp
 * @author Sebastian Oberschwendtner (sebastian.oberschwendtner@tum.de)
 * @brief Implements the CSV parser for deck values.
 * @version 3.0.0
 * @date 2024-02-19
 */

/* === Includes === */
#include "engine/engine_deck.h"
#include <algorithm>
#include <boost/multi_array.hpp>
#include <cmath>
#include <ranges>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>
#include <unitConversion/constants.h>

/* === Functions === */
auto DeckValue::is_within_range(const OperatingPoint &operating_point) const -> bool
{
    /* Check the three components */
    const bool N_within_range = operating_point.N >= data_.N.back() && operating_point.N <= data_.N.front();
    const bool Mach_within_range = operating_point.Mach >= data_.Mach.front() && operating_point.Mach <= data_.Mach.back();
    const bool FL_within_range = operating_point.altitude >= data_.FL.front() && operating_point.altitude <= data_.FL.back();

    /* Return the result */
    return (N_within_range && Mach_within_range && FL_within_range);
}

auto DeckValue::get_value_at(const OperatingPoint &operating_point) const -> double
{
    /* Check whether input is within the range of the deck value */
    if (!this->is_within_range(operating_point))
    {
        throw std::out_of_range("The operating point is out of range.");
    }

    /* Get the iterators of the data closest to the given operating point */
    const auto N_bound = std::ranges::lower_bound(data_.N, operating_point.N, std::greater{}); // The N values are in reverse order
    const auto N1 = N_bound == data_.N.cbegin() ? N_bound + 1 : N_bound;
    const auto N0 = N1 - 1;
    const auto Mach_bound = std::ranges::lower_bound(data_.Mach, operating_point.Mach);
    const auto Mach1 = Mach_bound == data_.Mach.cbegin() ? Mach_bound + 1 : Mach_bound;
    const auto Mach0 = Mach1 - 1;
    const auto FL_bound = std::ranges::lower_bound(data_.FL, operating_point.altitude);
    const auto FL1 = FL_bound == data_.FL.cbegin() ? FL_bound + 1 : FL_bound;
    const auto FL0 = FL1 - 1;

    /* Get the indices of the data in the value array */
    const auto index_N0 = std::distance(data_.N.cbegin(), N0);
    const auto index_N1 = index_N0 + 1;
    const auto index_Mach0 = std::distance(data_.Mach.cbegin(), Mach0);
    const auto index_Mach1 = index_Mach0 + 1;
    const auto index_FL0 = std::distance(data_.FL.cbegin(), FL0);
    const auto index_FL1 = index_FL0 + 1;

    /* Get the interpolation factors */
    const auto factor_N = (operating_point.N - *N0) / (*N1 - *N0);
    const auto factor_Mach = (operating_point.Mach - *Mach0) / (*Mach1 - *Mach0);
    const auto factor_FL = (operating_point.altitude - *FL0) / (*FL1 - *FL0);

    /* Get the corner values for the interpolation from the value data
     * => The index_range is a half-open range, meaning the last index
     * is not included in the result, therefore the range end is n+1.
     */
    using range = boost::multi_array_types::index_range;
    const auto boundary_indices =
        boost::indices[range(index_N0, index_N1 + 1)]
                      [range(index_FL0, index_FL1 + 1)]
                      [range(index_Mach0, index_Mach1 + 1)];
    const auto corner_points = data_.values[boundary_indices];

    /* === Trilinear interpolation === */
    /* Interpolate the N dimension */
    boost::multi_array<double, 2> N_interpolated(boost::extents[2][2]);
    for (std::size_t i = 0; i < 2; ++i)
    {
        for (std::size_t j = 0; j < 2; ++j)
        {
            N_interpolated[i][j] = corner_points[0][i][j] * (1 - factor_N) + corner_points[1][i][j] * factor_N;
        }
    }

    /* Interpolate the FL dimension */
    boost::multi_array<double, 1> FL_interpolated(boost::extents[2]);
    for (std::size_t i = 0; i < 2; ++i)
    {
        FL_interpolated[i] = N_interpolated[0][i] * (1 - factor_FL) + N_interpolated[1][i] * factor_FL;
    }

    /* Interpolate the Mach dimension */
    const auto value = FL_interpolated[0] * (1 - factor_Mach) + FL_interpolated[1] * factor_Mach;

    /* Return the result */
    return value;
}

auto DeckValue::lower_boundary() const -> OperatingPoint
{
    OperatingPoint minimal_op{};

    minimal_op.N = data_.N.back();
    minimal_op.Mach = data_.Mach.front();
    minimal_op.altitude = data_.FL.front();

    return minimal_op;
}

auto DeckValue::upper_boundary() const -> OperatingPoint
{
    OperatingPoint maximal_op{};

    maximal_op.N = data_.N.front();
    maximal_op.Mach = data_.Mach.back();
    maximal_op.altitude = data_.FL.back();

    return maximal_op;
}

//EngineDeck::EngineDeck(const std::string &filename)
//    : aName(filename)
//{
//    // myRuntimeInfo->debug << "Deck: "<< aName << std::endl;
//    /** Engine Performance Deck anlegen **/
//    std::string buffer;
//    std::ifstream is(aName.c_str());
//    if (!is)
//    {
//        myRuntimeInfo->debug << "Could not find deck: " << filename << std::endl;
//        *this = EngineDeck();
//    }
//    else
//    {
//        std::string row;
//        std::string delim = ";";
//        int i(0);
//        int j(0);
//        int k(0);
//        int n(0);
//        int m(0);
//        int p(0);
//        while (getline(is, row))
//        {
//            std::string::size_type lastPos = row.find_first_not_of(delim, 0); // Skip delimiters at beginning.
//            std::string::size_type pos = row.find_first_of(delim, lastPos);   // Find first "non-delimiter".
//            while (std::string::npos != pos || std::string::npos != lastPos)
//            {
//                buffer = row.substr(lastPos, pos - lastPos);
//                if (isalpha(buffer.c_str()[0]))
//                {
//                    buffer = "0"; // Ungueltige Werte mit Nullen besetzen
//                }
//                if (j == 0 && atof(buffer.c_str()) < 12 && atof(buffer.c_str()) > 0)
//                {
//                    N.push_back(atof(buffer.c_str()));
//                    k++; // Tabellenzaehler hochsetzen
//                }
//                if (i == 0)
//                {
//                    if (j > 0)
//                    {
//                        Mach.push_back(atof(buffer.c_str()));
//                    }
//                }
//                else
//                {
//                    if (j == 0 && k == 1)
//                    {
//                        FL.push_back(atof(buffer.c_str()));
//                    }
//                }
//                lastPos = row.find_first_not_of(delim, pos);
//                pos = row.find_first_of(delim, lastPos);
//                j++; // Spalte weiterzaehlen
//            }
//            i++;   // Zeile weiterzaehlen
//            j = 0; // Zaehler an Spaltenanfang
//        }
//        /** Maximal- und Minimalwerte speichern **/
//        FL_min = *min_element(FL.begin(), FL.end());
//        FL_max = *max_element(FL.begin(), FL.end());
//        Mach_min = *min_element(Mach.begin(), Mach.end());
//        Mach_max = *max_element(Mach.begin(), Mach.end());
//        N_min = *min_element(N.begin(), N.end());
//        N_max = *max_element(N.begin(), N.end());
//        /** Nun die eigentlichen Werte abholen **/
//        n = FL.size();
//        m = Mach.size();
//        p = N.size();
//        //        myRuntimeInfo->debug << n <<" x "<<m<<" x "<<p<< std::endl;
//        values = dtensor(n, m, p);
//        is.clear();
//        is.seekg(std::ios::beg);
//        is.setf(std::ios::skipws);
//        is.precision(10);
//        i = 0;
//        j = 0;
//        k = 0;
//        while (getline(is, row))
//        {
//            if (i == n + 1)
//            {
//                i = 0;
//                k++;
//            }
//            if (k == p)
//                break;
//            std::string::size_type delimPos1(0);                                    // Zeiger an der Anfang.
//            std::string::size_type delimPos2 = row.find_first_of(delim, delimPos1); // Find first "Delimiter".
//            while (std::string::npos != delimPos1 && std::string::npos != delimPos2)
//            {
//                buffer = row.substr(delimPos1, delimPos2 - delimPos1);
//                if (isalpha(buffer.c_str()[0]))
//                {
//                    buffer = "0"; // Ungueltige Werte mit Nullen besetzen
//                }
//                if (i > 0 && j > 0)
//                {
//                    values[i - 1][j - 1][k] = atof(buffer.c_str());
//                }
//                delimPos1 = row.find_first_not_of(delim, delimPos2);
//                delimPos2 = row.find_first_of(delim, delimPos1);
//                j++; // Spalte weiterzaehlen
//            }
//            i++;   // Zeile weiterzaehlen
//            j = 0; // Zaehler an Spaltenanfang
//        }
//    }
//    is.close();
//    // cout << "check" << std::endl;
//}
//
//EngineDeck::EngineDeck() : FL_min(0),
//                           FL_max(0),
//                           Mach_min(0),
//                           Mach_max(0),
//                           N_min(0),
//                           N_max(0),
//                           aName("")
//{
//    values = dtensor(0, 0, 0);
//}
//
//EngineDeck::~EngineDeck()
//{
//}
//
//double EngineDeck::getVal(double *FL_mom, double *Mach_mom, double *N_mom, const bool &checkValidValues, const bool &changeSpoolSpeed) const
//{
//    if (aName == "")
//        return NAN;
//    /* Interpolation im Raume */
//    int pos[3][2] = {{0}};               // Stellen fuer das Interpolieren
//    double hebel[3] = {0};               // Interpolationshebel
//    std::vector<double> VTemp;           // std::vector of deltas between all values in deck file and passed values
//    VTemp.reserve(1000);                 // minimize reallocations of the std::vector to save time
//    std::vector<double>::iterator itMin; // iterator pointing to the smallest delta
//    size_t foundPos(0);                  // position of the smallest delta at std::vector
//    // if passed values are higher/lower than available values in deck file, take the maximum/minimum values
//    if (*FL_mom < *min_element(this->FL.begin(), this->FL.end()))
//        *FL_mom = *min_element(this->FL.begin(), this->FL.end());
//    if (*FL_mom > *max_element(this->FL.begin(), this->FL.end()))
//        *FL_mom = *max_element(this->FL.begin(), this->FL.end());
//    if (*Mach_mom < *min_element(this->Mach.begin(), this->Mach.end()))
//        *Mach_mom = *min_element(this->Mach.begin(), this->Mach.end());
//    if (*Mach_mom > *max_element(this->Mach.begin(), this->Mach.end()))
//        *Mach_mom = *max_element(this->Mach.begin(), this->Mach.end());
//    if (std::isnan(*N_mom) || *N_mom < *min_element(this->N.begin(), N.end()))
//        *N_mom = *min_element(this->N.begin(), this->N.end());
//    if (*N_mom > *max_element(this->N.begin(), N.end()))
//        *N_mom = *max_element(this->N.begin(), this->N.end());
//    /** FL Interpolationsstellen **/
//    for (unsigned int i(0); i < this->FL.size(); i++)
//        VTemp.push_back(fabs(this->FL.at(i) - *FL_mom)); // push back deltas into Vtemp
//    itMin = min_element(VTemp.begin(), VTemp.end());     // find smallest delta
//    foundPos = static_cast<size_t>(itMin - VTemp.begin());
//    // cout<<"Pos: " << foundPos << std::endl;
//    if (foundPos == 0)
//    { // first position chosen
//        pos[0][0] = foundPos;
//        pos[0][1] = foundPos + 1;
//    }
//    else if (foundPos < this->FL.size() - 1 && inRange(this->FL.at(foundPos), this->FL.at(foundPos + 1), *FL_mom))
//    {
//        pos[0][0] = foundPos;
//        pos[0][1] = foundPos + 1;
//    }
//    else
//    { // no bigger data available or delta smaller than the closest value
//        pos[0][0] = foundPos - 1;
//        pos[0][1] = foundPos;
//    }
//    VTemp.clear();
//    /** Mach Interpolationsstellen **/
//    for (unsigned int i(0); i < this->Mach.size(); i++)
//        VTemp.push_back(fabs(this->Mach.at(i) - *Mach_mom));
//    itMin = min_element(VTemp.begin(), VTemp.end());
//    foundPos = static_cast<size_t>(itMin - VTemp.begin());
//    // cout<<"Pos: " << foundPos << std::endl;
//    if (foundPos == 0)
//    {
//        pos[1][0] = foundPos;
//        pos[1][1] = foundPos + 1;
//    }
//    else if (foundPos < this->Mach.size() - 1 && inRange(this->Mach.at(foundPos), this->Mach.at(foundPos + 1), *Mach_mom))
//    {
//        pos[1][0] = foundPos;
//        pos[1][1] = foundPos + 1;
//    }
//    else
//    {
//        pos[1][0] = foundPos - 1;
//        pos[1][1] = foundPos;
//    }
//    VTemp.clear();
//    /** Schalter fuer Sonderfaelle **/
//    bool interpolSucceded(false);
//    bool extrapolate(false);
//    bool changeInterpolatePos(false);
//    // bool repaired(false);
//    int tries(0);
//    double val(0);
//    /** Interpolation des Wertes aus Deck **/
//    do
//    {
//        /** N Interpolationsstellen **/
//        VTemp.clear();
//        for (unsigned int i(0); i < this->N.size(); i++)
//            VTemp.push_back(fabs(this->N.at(i) - *N_mom)); //
//        itMin = min_element(VTemp.begin(), VTemp.end());
//        foundPos = static_cast<size_t>(itMin - VTemp.begin());
//        if (changeSpoolSpeed && (extrapolate || changeInterpolatePos))
//        {
//            // if interpolation or extrapolation needed and most critical value=-9999, then increase LPC in order to find a more stable engine region
//            bool validFound = false;
//            bool highestSpoolSpeedReached = false;
//            bool lowestSpoolSpeedReached = false;
//            unsigned int it = 0;
//            do
//            {
//                pos[2][0] = foundPos;
//                pos[2][1] = foundPos;
//                if (this->values[pos[0][0]][pos[1][1]][foundPos] != -9999 && this->values[pos[0][0]][pos[1][0]][foundPos] != -9999 && this->values[pos[0][1]][pos[1][1]][foundPos] != -9999 && this->values[pos[0][1]][pos[1][0]][foundPos] != -9999)
//                { // most critical position. Why? LPC small, high Mach number and low altitude
//                    validFound = true;
//                }
//                else
//                {
//                    if (foundPos == 0)
//                    {
//                        highestSpoolSpeedReached = true;
//                    }
//                    if (foundPos == this->N.size() - 1)
//                    {
//                        lowestSpoolSpeedReached = true;
//                    }
//                    if (highestSpoolSpeedReached == false)
//                    {
//                        foundPos--; // increase LPC speed in order to find more stable engine data
//                    }
//                    else if (lowestSpoolSpeedReached == false)
//                    {
//                        foundPos++; // decrease spool speed in order to find more stable engine data (low altitudes, high Mach number)
//                    }
//                    else
//                    {
//                        std::stringstream errorMessage;
//                        errorMessage << "No valid values found in EngineDeck::getVal for: " << std::endl;
//                        errorMessage << "*FL_mom: " << *FL_mom << std::endl;
//                        errorMessage << "*Mach_mom: " << *Mach_mom << std::endl;
//                        errorMessage << "*N_mom: " << *N_mom << std::endl;
//                        errorMessage << "Abort program." << std::endl;
//                        throwError(__FILE__, __func__, __LINE__, errorMessage.str());
//                    }
//                    it++;
//                }
//            } while (!validFound && (foundPos > 1) && ((this->N.at(foundPos) - *N_mom) < 0.2) && it < this->N.size()); // maximum allowable LPC speed 20%
//            *N_mom = this->N.at(foundPos);
//        }
//        else if (extrapolate || changeInterpolatePos)
//        {
//            pos[2][0] = foundPos - 1;
//            pos[2][1] = foundPos;
//        }
//        /*else if (changeInterpolatePos)
//        {
//            pos[2][0] = foundPos;
//            pos[2][1] = foundPos+1;//
//        }*/
//        // foundPos+1; changed this because at lower LPC speeds the possibility to find valid values is lower
//        else
//        { // different to altitude and Mach number because N is sorted descendent
//            if (foundPos == 0)
//            {
//                pos[2][0] = foundPos + 1;
//                pos[2][1] = foundPos;
//            }
//            else if (foundPos < this->N.size() - 1 &&
//                     inRange(this->N.at(foundPos + 1), this->N.at(foundPos), *N_mom))
//            {
//                pos[2][0] = foundPos + 1;
//                pos[2][1] = foundPos;
//            }
//            else
//            {
//                pos[2][0] = foundPos;
//                pos[2][1] = foundPos - 1;
//            }
//        }
//        /** Hebel bestimmen **/
//        /** Reihenfolge: FL, Mach, N **/
//        hebel[0] = (*FL_mom - this->FL.at(pos[0][0])) / (this->FL.at(pos[0][1]) - this->FL.at(pos[0][0]));
//        hebel[1] = (*Mach_mom - this->Mach.at(pos[1][0])) / (this->Mach.at(pos[1][1]) - this->Mach.at(pos[1][0]));
//        hebel[2] = (*N_mom - this->N.at(pos[2][0])) / (this->N.at(pos[2][1]) - this->N.at(pos[2][0]));
//        /** Sonderfall keine Interpolation zwischen N1 noetig **/
//        if (checkValidValues)
//        {
//            if (fabs(*N_mom - this->N.at(pos[2][0])) < ACCURACY_MEDIUM)
//            {
//                pos[2][1] = pos[2][0];
//                hebel[2] = 0.;
//            }
//            else if (fabs(*N_mom - this->N.at(pos[2][1])) < ACCURACY_MEDIUM)
//            {
//                pos[2][0] = pos[2][1];
//                hebel[2] = 0.;
//            }
//        }
//        /** Interpolieren **/
//        double v1 = this->values[pos[0][0]][pos[1][0]][pos[2][0]]; // lower N1, lower altitude, lower mach
//        double v2 = this->values[pos[0][0]][pos[1][0]][pos[2][1]]; // higher N1, lower altitude, lower mach
//        double v3 = this->values[pos[0][1]][pos[1][0]][pos[2][0]]; // lower N1, higher altitude, lower mach
//        double v4 = this->values[pos[0][1]][pos[1][0]][pos[2][1]]; // higher N1, higher altitude, lower mach
//        double v5 = this->values[pos[0][0]][pos[1][1]][pos[2][0]]; // lower N1, lower altitude, higher mach
//        double v6 = this->values[pos[0][0]][pos[1][1]][pos[2][1]]; // higher N1, lower altitude, higher mach
//        double v7 = this->values[pos[0][1]][pos[1][1]][pos[2][0]]; // lower N1, higher altitude, higher mach
//        double v8 = this->values[pos[0][1]][pos[1][1]][pos[2][1]]; // higher N1, higher altitude, higher mach
//        /** Schalter fuer ungueltige Eintraege setzen **/
//        bool invalidValueInFstPos = (v1 == -9999. || v3 == -9999. || v5 == -9999. || v7 == -9999.); // invalid values for initial LPC speed
//        bool invalidValueInScdPos = (v2 == -9999. || v4 == -9999. || v6 == -9999. || v8 == -9999.); // invalid values for final LPC speed
//        /** Fallunterscheidung fuer Interpolation **/
//        if (checkValidValues)
//        {
//            if (invalidValueInFstPos && invalidValueInScdPos && !changeSpoolSpeed)
//            { // only if increaseLpcSpeed is false
//                myRuntimeInfo->err << "Error with interpolation in engine decks. At least 1 invalid value in both N1 decks!" << std::endl;
//                myRuntimeInfo->err << "Deckname: " << this->aName << std::endl;
//                myRuntimeInfo->err << "N1_1 = " << N.at(pos[2][0]) << ", N1_2 = " << N.at(pos[2][1]) << std::endl;
//                myRuntimeInfo->err << "Target values for interpolation: Alt=" << *FL_mom << " m, Ma=" << *Mach_mom << ", N1=" << *N_mom << std::endl;
//                myRuntimeInfo->err << "Values at all grid points: " << v1 << " | " << v2 << " | " << v3 << " | " << v4 << " | "
//                                   << v5 << " | " << v6 << " | " << v7 << " | " << v8 << std::endl;
//                throwError(__FILE__, __func__, __LINE__, "Invalid interpolation value in engine deck. Abort program!");
//            }
//            else if (invalidValueInFstPos && invalidValueInScdPos && changeSpoolSpeed)
//            {
//                myRuntimeInfo->debug << "Change Spool speed due to invalid interpolation values..." << std::endl;
//                changeInterpolatePos = true;
//            }
//            else if (!invalidValueInFstPos && invalidValueInScdPos)
//            {
//                myRuntimeInfo->debug << "Extrapolation in engine deck ..." << std::endl;
//                extrapolate = true;
//            }
//            else if (invalidValueInFstPos && !invalidValueInScdPos)
//            {
//                myRuntimeInfo->debug << "Extended interpolation in engine deck ..." << std::endl;
//                changeInterpolatePos = true;
//            }
//            else
//            { // success
//                if (extrapolate)
//                {
//                    myRuntimeInfo->debug << "... succeeded" << std::endl;
//                    extrapolate = false;
//                    // repaired = true;
//                }
//                if (changeInterpolatePos)
//                {
//                    myRuntimeInfo->debug << "... succeeded" << std::endl;
//                    changeInterpolatePos = false;
//                    // repaired = true;
//                }
//                interpolSucceded = true;
//            }
//        }
//        else
//        {
//            interpolSucceded = true;
//        }
//        /** Interpolation **/
//        val = ((v1 + hebel[2] * (v2 - v1)) + hebel[1] * ((v5 + hebel[2] * (v6 - v5)) - (v1 + hebel[2] * (v2 - v1)))) +
//              hebel[0] * (((v3 + hebel[2] * (v4 - v3)) + hebel[1] * ((v7 + hebel[2] * (v8 - v7)) - (v3 + hebel[2] * (v4 - v3)))) -
//                          ((v1 + hebel[2] * (v2 - v1)) + hebel[1] * ((v5 + hebel[2] * (v6 - v5)) - (v1 + hebel[2] * (v2 - v1)))));
//        //        myRuntimeInfo->debug << "**** INTERPOLATION: ****" << std::endl;
//        //        myRuntimeInfo->debug << "    - v1:    " << v1 << std::endl;
//        //        myRuntimeInfo->debug << "    - v2:    " << v2 << std::endl;
//        //        myRuntimeInfo->debug << "    - v3:    " << v3 << std::endl;
//        //        myRuntimeInfo->debug << "    - v4:    " << v4 << std::endl;
//        //        myRuntimeInfo->debug << "    - v5:    " << v5 << std::endl;
//        //        myRuntimeInfo->debug << "    - v6:    " << v6 << std::endl;
//        //        myRuntimeInfo->debug << "    - v7:    " << v7 << std::endl;
//        //        myRuntimeInfo->debug << "    - v8:    " << v8 << std::endl;
//        //        myRuntimeInfo->debug << "--------------------"<< std::endl;
//        //        myRuntimeInfo->debug << "    - val:    " << val << std::endl;
//        //        myRuntimeInfo->debug << "**** INTERPOLATION end ****" << std::endl;
//        /** Abfangen, ob Extrapolation oder erweiterte Interpolation evtl. fehlgeschlagen ist **/
//        tries++;
//        if (tries > 4)
//        {
//            std::stringstream errorMessage;
//            errorMessage << "Error with interpolation in engine decks. Too many trials for special case treatment!" << std::endl;
//            errorMessage << "Deckname: " << this->aName << std::endl;
//            errorMessage << "N1_1 = " << N.at(pos[2][0]) << ", N1_2 = " << N.at(pos[2][1]) << std::endl;
//            errorMessage << "Target values for interpolation: Alt=" << *FL_mom << " m, Ma=" << *Mach_mom << ", N1=" << *N_mom << std::endl;
//            errorMessage << "Values at all grid points: " << v1 << " | " << v2 << " | " << v3 << " | " << v4 << " | " << v5 << " | " << v6 << " | " << v7 << " | " << v8
//                         << std::endl;
//            errorMessage << "Temp. interpolation value: " << val << std::endl;
//            errorMessage << "Switch extrapolation: " << ((extrapolate) ? "enabled" : "disabled") << std::endl;
//            errorMessage << "Switch extended interpolation: " << ((changeInterpolatePos) ? "enabled" : "disabled") << std::endl;
//            errorMessage << "Abort program!" << std::endl;
//            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
//        }
//    } while (!interpolSucceded);
//    return val;
//}
//
//std::vector<std::vector<std::vector<double>>> EngineDeck::dtensor(const int &n, const int &m, const int &p)
//{
//    /** n: zeilen, m: spalten, p: tiefen **/
//    // allocation
//    std::vector<std::vector<std::vector<double>>> theValueMatrix;
//    theValueMatrix.resize(n);
//    for (std::vector<std::vector<double>> &aValueMatrix : theValueMatrix)
//    {
//        aValueMatrix.resize(m);
//        for (std::vector<double> &aVector : aValueMatrix)
//        {
//            aVector.resize(p);
//        }
//    }
//    return theValueMatrix;
//}
