/*  Copyright (C) 2019-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef STANDARDFILES_INCLUDE_STANDARDFILES_OUTPUT_H_
#define STANDARDFILES_INCLUDE_STANDARDFILES_OUTPUT_H_

#include <cstdint>
#include <string>
#include <fstream>
#include <sstream>
#include <vector>
#include <numeric>

#ifdef BUILD_STANDARDFILES_SHARED
    #ifdef _WIN32
        #define STANDARDFILESDLLEXPORT __declspec(dllexport)
    #else
        #define STANDARDFILESDLLEXPORT __attribute__ ((visibility ("default")))
    #endif
#elif defined(IMPORT_STANDARDFILES_SHARED)
    #ifdef _WIN32
        #define STANDARDFILESDLLEXPORT __declspec(dllimport)
    #else
        #define STANDARDFILESDLLEXPORT
    #endif
#else
    #define STANDARDFILESDLLEXPORT
#endif

#ifndef USE_GNUPLOT
    #define USE_GNUPLOT 0
#endif

/**
*  Abstract class to generate and manage the program outputs
**/

class STANDARDFILESDLLEXPORT output {
 public:
    //Constructor and Destructor
    /** \brief Constructor of the output object
    *   \param programName name of the current program
    *   \param programVersion version of the current program
    *   \param aircraftName name of the current aircraft
    *   \param ioDirectory path to the Input-/Output-Directory
    */
    output(const std::string& programName, const std::string& programVersion, const std::string& aircraftName, const std::string& ioDirectory);

    /** \brief Destructor of the output object
    */
    virtual ~output();

    //Member functions
    /** \brief Function initializes variables for output and plotting
    *   \details function should invoke initialzeOutputSettings to set output-switches and paths
    */
    virtual void initialize() = 0;

    /** \brief Function is called from program to write the output
    *   \details output for reporting (HTML/TeX) and information (info/guiSettings) as well as plots are written depending on the output settings
    */
    void writeOutput();

 protected:
    //Structs
    /** \struct plotArguments holds the arguments for plotting output
    */
    struct plotArguments {
        std::string name;                                   /**< name of the plot */
        std::string plotScriptFilename;  /**< name of the plotscript; if DEFAULT_PLOTFILE, the plotfile with the plot name is used */
        std::string outputName;                   /**< output name of the plot; if DEFAULT, the name of the plot is used */
        std::string additionalArguments;                 /**< string contains additional gnuplot args for plot scripts that have special external variables */
        STANDARDFILESDLLEXPORT explicit plotArguments(const std::string& plotName, const std::string& plotScriptName = "DEFAULT", const std::string& outName = "DEFAULT",
                                                      const std::string& addArgs = "");
    };

    //Member functions
    /** \brief Initializer for internal struct outputSettings
    *   \details function initialize all relevant output variables stored in theOutputSettings
    *   \param writeXmlOutput   switch to enable/disable XML Output into AircraftExchangeFile
    *   \param generatePlots    switch to enable/disable plot generation
    *   \param copyPlottingFiles switch to enable/disable copying CSV and PLT files to project folder
    *   \param deletePlottingFilesTool switch to enable/disable deleting CSV and PLT folders
    *   \param writeHtml        switch to enable/disable generation of HTML report
    *   \param writeTex         switch to enable/disable generation of TeX report
    *   \param writeInfoFiles   switch to enable/disable generation of information files (<tool>_info.txt and <tool>_guiSettings.xml)
    *   \param useGnuplot       switch to enable/disable gnuplot for svg plots during compilation time
    *   \param aPlotDirectory   string of full plot directory path; if DEFAULT, the default plot directory 'plots' in the input-/output-directory is used
    *   \param aHtmlReportDirectory   string of full HTML-report directory path; if DEFAULT, the default directory 'reportHTML' in the input-/output-directory is used
    *   \param aTexReportDirectory    string of full TeX-report directory path; if DEFAULT, the default directory 'reportTex' in the input-/output-directory is used
    *   \param gnuplotPath      string of full path to gnuplot directory; if DEFAULT, the default setting is used, where the gnuplot program is one level above the running program
    *   \param inkscapePath     string of full path to inkscape directory; if DEFAULT, the default setting is used, where the inkscape program is one level above the running program
    */
    void initializeOutputSettings(const bool& writeXmlOutput, const bool& generatePlots, const bool& copyPlottingFiles, const bool& deletePlottingFilesTool,
                                  const bool& writeHtml, const bool& writeTex,
                                  const bool& writeInfoFiles, const bool& useGnuplot = USE_GNUPLOT,
                                  const std::string& aPlotDirectory = "DEFAULT", const std::string& aHtmlReportDirectory = "DEFAULT",
                                  const std::string& aTexReportDirectory = "DEFAULT", const std::string& gnuplotPath = "DEFAULT", const std::string& inkscapePath = "DEFAULT");

    /** \brief Function writes the changes data into Aircraft-XML
    *   \details if nothing has to be written, only an informative output is given
    */
    virtual void writeXmlOutput() = 0;

    /** \brief Function generates the body of HTML report
    *   \details the function is full virtual and is implemented in each program; it defines the content BETWEEN the <body> tag of the HTML-report
    *   \return report a stringstream of the information for the body-tag of HTML-Report
    */
    virtual std::stringstream generateHtmlReportBody() = 0;

    /** \brief Function generates a section for the TeX report
    *   \details the function is full virtual and is implemented in each program; it defines the content of a section of the TeX-report
    *   \return report a stringstream of the information for the TeX-Report
    */
    virtual std::stringstream generateTexReportBody() = 0;

    #if USE_GNUPLOT
    /** \brief Function generates all data for the plots in plots-Vector
    *   \details the function is full virtual and is implemented in each program; often sub-methods are called which define program specific data generation
    */
    virtual void generatePlotData() = 0;

    /** \brief Function generates all plot scripts for the plots in plots-Vector
    *   \details the function is full virtual and is implemented in each program; often sub-methods are called which define program specific plot script generation
    */
    virtual void generatePlotScripts() = 0;

    /** \brief Function generates a SVG plot using the svgPlot library
    *   \details in this case it is used as placeholder
    */
    virtual void generateSvgPlot() {}
    #else
    /** \brief Function generates all data for the plots in plots-Vector
    *   \details the function is full virtual and is implemented in each program; often sub-methods are called which define program specific data generation
    */
    virtual void generatePlotData() {}

    /** \brief Function generates all plot scripts for the plots in plots-Vector
    *   \details the function is full virtual and is implemented in each program; often sub-methods are called which define program specific plot script generation
    */
    virtual void generatePlotScripts() {}

    /** \brief Function generates a SVG plot using the svgPlot library
    */
    virtual void generateSvgPlot() {}
    #endif

    /** \brief Function generates standardized header of the plot script
    *   \param script: ofstream that contains the script content
    *   \param format: determines height and width of the plot
    *   \param delimiter: separator for the data (e.g. ";")
    *   \param xtraLineStyles: bool to activate additional line styles
    *   \param doubleSize: bool to double the size of the plot
    */
    void generatePlotScriptHeader(std::ofstream &script, const std::string& format, const std::string& delimiter, const bool& xtraLineStyles, const bool& doubleSize);

    /** \brief Function to convert all svg files in the plot directory to pdf files
    *   \details Function calls convertSVG2PDF
    *   \param aDirectory: full path to the plot directory; if GLOBAL_PLOTDIR is set, the directory defined in theDefaultDirectories.plotDirectory is used
    */
    void convertAllSVG2PDF(std::string aDirectory = "GLOBAL_PLOTDIR");

    //Member variables
    std::vector<plotArguments> plots; /**< vector contains names of all plots to be generated */

 private:
    /** \struct defaultDirectories contains the paths to the default output directories
    *   \details the defaultDirectories are located in the input-/output-directory
    */
    const struct defaultDirectories {
        std::string plotDirectory;  /**< path to the default plot-directory ioDir/reporting/plots/ */
        std::string toolCsvDirectory; /**< path to csv directory of each tool */
        std::string csvDirectory; /**< path to the default csv directory ioDir/reporting/plots/csvFiles */
        std::string toolPltDirectory; /**< path to plt directory of each tool */
        std::string pltDirectory; /**< path to the default plt directory ioDir/reporting/plots/pltFiles */
        std::string htmlDirectory;  /**< path to the default HTML-directory ioDir/reporting/reportHTML/ */
        std::string texDirectory;   /**< path to the default TeX-directory ioDir/reporting/reportTeX/ */
        std::string gnuplotPath;    /**< path to the default gnuplot-directory ../gnuplot/gnuplot.exe */
        std::string inkscapePath;   /**< path to the default inkscape-directory ../inkscape/inkscape.exe */
    } theDefaultDirectories; /**< Variable of type defaultDirectories for access to corresponding struct members */

    /** \struct outputSettings contains the settings for the output
    *   \details the outputSettings are set in initializeOutputSettings and are the controls over which output is disabled/enabled
    */
    struct outputSettings {
        bool writeXmlOutput;    /**< switch to enable/disable XML Output into AircraftExchangeFile */
        bool generatePlots;     /**< switch to enable/disable plot generation */
        bool copyPlottingFiles; /**< switch to enable/disable copying CSV and PLT files to project folder */
        bool deletePlottingFilesTool; /**< switch to enable/disable deleting CSV and PLT folders in tool*/
        bool writeHtml;         /**< switch to enable/disable generation of HTML report */
        bool writeTex;          /**< switch to enable/disable generation of TeX report */
        bool writeInfofiles;    /**< switch to enable/disable generation of information files (<tool>_info.txt and <tool>_guiSettings.xml) */
        bool useGnuplot;        /**< switch to enable/disable gnuplot for svg plots during compilation time */

        std::string thePlotDirectory;       /**< string of full plot directory path; if DEFAULT, the default plot directory 'plots' in the input-/output-directory is used */
        std::string theHtmlReportDirectory; /**< string of full HTML-report directory path; if DEFAULT, the default directory 'reportHTML' in the input-/output-directory is used */
        std::string theTexReportDirectory;  /**< string of full TeX-report directory path; if DEFAULT, the default directory 'reportTex' in the input-/output-directory is used */
        std::string
        inkscapePath;  /**< string of full path to inkscape directory; if DEFAULT, the default setting is used, where the inkscape program is one level above the running program */
        std::string
        gnuplotPath;   /**< string of full path to gnuplot directory; if DEFAULT, the default setting is used, where the gnuplot program is one level above the running program */
    }theOutputSettings; /**< Variable of type outputSettings for access to corresponding struct members */

    //Member functions
    /** \brief function checks if the given paths in theOutputSettings exist
    *   \details if the path given in theOutputSettings does not exist, it is created except for gnuplot and inkscape
    */
    void checkOutputPaths();

    /** \brief function generates plots for all elements of the vector 'plots'
    *   \details function first writes a file for each plot-element containing the data for plotting, then a plot script for each element and finally gnuplot is executed with the written scripts
    */
    void generatePlot();

    /** \brief function copies all CSV and PLT files to the project folder
     * \details (To be added)
     */
    void copyPlottingFiles();

    /** \brief function deletes all CSV and PLT files from tool folder
     * \details Tool generates plots and csv files in folders which lie in the tool folder
     * These folders are also copied to the projects folder if the switch copyPlottingFiles is switched on
     */
    void deletePlottingFiles();

    /** \brief function creates all plots given in vector 'plots' with gnuplot and converts the SVG plots into PDF if 'writeTeX' is true
    */
    void createAndConvertPlot();

    /** \brief function executes gnuplot with a plotFile
    *   \param gnuplotArgs name, plotScript and additional arguments for plot generated with gnuplot
    */
    void runGnuplot(plotArguments gnuplotArgs);

    /** \brief Function to convert svg file to pdf file
    *   \param aFile: name of the file to be converted without path
    */
    void convertSVG2PDF(const std::string& aFile);

    /** \brief Function generates the info text file
    *   \details Function creates a info file and writes all the input and output xml-paths of the aircraft xml
    */
    void generateInfoFiles();

    /** \brief Function generates the _input and _output xml-files
    *   \details In tree structure with attributes and values of the respective read and written files
    *   \param variablesVector: Either the node::input or node::output vector are read for input/output
    *   \param direction: String to identify whether the file to be written is _input or _output
    */
    void generateInputOutputXMLFile(std::vector<std::string> variablesVector, std::string direction);

    /** \brief Function generates the gui settings xml-file
    *   \details same file as info text file, but formatted in xml
    */
    void generateGuiSettingsFile();

    /** \brief Function generates the HTML-report
    *   \details opens an html file and calls the functions generateHtmlReportHeader/generateHtmlReportBody/generateHtmlReportFooter
    */
    void generateHtmlReport();

    /** \brief Function generates the header for the html report
    *   \return report stringstream that contains the content of the header
    */
    std::stringstream generateHtmlReportHeader();

    /** \brief Function generates the footer for the html report
    *   \return report stringstream that contains the content of the footer
    */
    std::stringstream generateHtmlReportFooter();

    /** \brief Function generates the TeX-report
    *   \details opens an tex file and calls the function generateTeXReportBody
    */
    void generateTexReport();

    //Member variables
    std::string mProgramName;   /**< Name of the current running program */
    std::string programVersion; /**< Version number of the current running program */
    std::string theAircraft;    /**< Name of the current running program */

    double columnWidth; /**< Width of the images in the HTML report in pixels */
    double columnSpace; /**< Column space in the HTML report in pixels */
};
#endif // STANDARDFILES_INCLUDE_STANDARDFILES_OUTPUT_H_
