/*  Copyright (C) 2008-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "standardFiles/output.h"
#ifdef _WIN32
    #include <windows.h>
#endif
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include "standardFiles/functions.h"

output::output(const std::string& programName, const std::string& programVersion, const std::string& aircraftName, const std::string& ioDirectory)
    :
    theDefaultDirectories({.plotDirectory = ioDirectory + "reporting" + FILESEPERATOR + "plots" + FILESEPERATOR,
                           .toolCsvDirectory = std::string("csvFiles") + FILESEPERATOR,
                           .csvDirectory = ioDirectory + "reporting" + FILESEPERATOR + "plots" + FILESEPERATOR + "csvFiles" + FILESEPERATOR,
                           .toolPltDirectory = std::string("pltFiles") + FILESEPERATOR,
                           .pltDirectory = ioDirectory + "reporting" + FILESEPERATOR + "plots" + FILESEPERATOR + "pltFiles" + FILESEPERATOR,
                           .htmlDirectory = ioDirectory + "reporting" + FILESEPERATOR + "reportHTML" + FILESEPERATOR,
                           .texDirectory  = ioDirectory + "reporting" + FILESEPERATOR + "reportTeX" + FILESEPERATOR,
                           .gnuplotPath   = GNUPLOTDEFAULTDIR,
                           .inkscapePath  = INKSCAPEDEFAULTDIR}),
    theOutputSettings({false, false, false, false, false, false, false, USE_GNUPLOT, theDefaultDirectories.plotDirectory, theDefaultDirectories.htmlDirectory,
                     theDefaultDirectories.texDirectory, theDefaultDirectories.inkscapePath, theDefaultDirectories.gnuplotPath}),
    mProgramName(programName),
    programVersion(programVersion),
    theAircraft(aircraftName),
    columnWidth(500.),
    columnSpace(24.) {
    //ctor
}

output::~output() {
    //dtor
}

// cppcheck-suppress unusedFunction
void output::initializeOutputSettings(const bool& writeXmlOutput, const bool& generatePlots, const bool& copyPlottingFiles, const bool& deletePlottingFilesTool,
                                      const bool& writeHtml, const bool& writeTex,
                                      const bool& writeInfoFiles, const bool& useGnuplot,
                                      const std::string& aPlotDirectory, const std::string& aHtmlReportDirectory, const std::string& aTexReportDirectory,
                                      const std::string& gnuplotPath, const std::string& inkscapePath) {
    //Switches
    this->theOutputSettings.writeXmlOutput = writeXmlOutput;
    this->theOutputSettings.generatePlots = generatePlots;
    this->theOutputSettings.copyPlottingFiles = copyPlottingFiles;
    this->theOutputSettings.deletePlottingFilesTool = deletePlottingFilesTool;
    this->theOutputSettings.writeHtml = writeHtml;
    this->theOutputSettings.writeTex = writeTex;
    this->theOutputSettings.writeInfofiles = writeInfoFiles;
    this->theOutputSettings.useGnuplot = useGnuplot;
    //Directories
    if (aPlotDirectory != "DEFAULT")
        this->theOutputSettings.thePlotDirectory = aPlotDirectory;
    if (aHtmlReportDirectory != "DEFAULT")
        this->theOutputSettings.theHtmlReportDirectory = aHtmlReportDirectory;
    if (aTexReportDirectory != "DEFAULT")
        this->theOutputSettings.theTexReportDirectory = aTexReportDirectory;
    if (gnuplotPath != "DEFAULT")
        this->theOutputSettings.gnuplotPath = gnuplotPath;
    if (inkscapePath != "DEFAULT")
        this->theOutputSettings.inkscapePath = inkscapePath;
    this->checkOutputPaths();
}

void output::writeOutput() { // cppcheck-suppress unusedFunction
    if (this->theOutputSettings.writeInfofiles) {
        myRuntimeInfo->out << "Write Input-Files ..." << std::endl;
        if (!node::input.empty()) {
            try {
                this->generateInputOutputXMLFile(node::input, "input");
            } catch (const std::exception& except) {
                myRuntimeInfo->err << except.what() << std::endl;
                myRuntimeInfo->warn << "No _input.xml file will be written." << std::endl;
            }
        }
    }
    if (this->theOutputSettings.writeXmlOutput) {
        myRuntimeInfo->out << "Write XML-Output ..." << std::endl;
        this->writeXmlOutput();
    }
    if (this->theOutputSettings.generatePlots) {
        myRuntimeInfo->out << "Create Plots ..." << std::endl;
        this->generatePlot();
    }
    if (this->theOutputSettings.generatePlots && this->theOutputSettings.copyPlottingFiles) {
        myRuntimeInfo->out << "Copy Plotting files ..." << std::endl;
        this->copyPlottingFiles();
    }
    if (this->theOutputSettings.writeHtml) {
        myRuntimeInfo->out << "Write HTML-Report ..." << std::endl;
        this->generateHtmlReport();
    }
    if (this->theOutputSettings.writeTex) {
        myRuntimeInfo->out << "Write TeX-Report ..." << std::endl;
        this->generateTexReport();
    }
    if (this->theOutputSettings.writeInfofiles) {
        myRuntimeInfo->out << "Write Output-Files ..." << std::endl;
        this->generateInfoFiles();
        this->generateGuiSettingsFile();
        if (!node::output.empty()) {
            try {
                this->generateInputOutputXMLFile(node::output, "output");
            } catch (const std::exception& except) {
                myRuntimeInfo->err << except.what() << std::endl;
                myRuntimeInfo->warn << "No _output.xml file will be written." << std::endl;
            }
        }
    }
    if (this->theOutputSettings.generatePlots && this->theOutputSettings.deletePlottingFilesTool) {
        myRuntimeInfo->out << "Delete Plotting files from tool ..." << std::endl;
        this->deletePlottingFiles();
    }
}

void output::checkOutputPaths() {
    //Plots
    if (this->theOutputSettings.generatePlots) {
        if (!fileExists(this->theOutputSettings.thePlotDirectory)) {
            myRuntimeInfo->warn << "The given desired output directory for plots '" << this->theOutputSettings.thePlotDirectory << "' does not exist." << std::endl;
            myRuntimeInfo->out << "Create folder for plots in given path: " << this->theOutputSettings.thePlotDirectory << std::endl;
            CreateDirectory(this->theOutputSettings.thePlotDirectory);
        }
        /* Check path for plotting programs */
        if (!fileExists(this->theOutputSettings.gnuplotPath)) {
            myRuntimeInfo->err << "Gnuplot-Path \"" << this->theOutputSettings.gnuplotPath << "\" could not be found. Change settings in config file. Abort program." << std::endl;
            exit(1);
        }
        if (!fileExists(this->theOutputSettings.inkscapePath) && this->theOutputSettings.writeTex) {
            myRuntimeInfo->err << "Inkscape-Path \"" << this->theOutputSettings.inkscapePath << "\" could not be found."
                               << " Change settings in config file. Abort program." << std::endl;
            exit(1);
        }
        if (!fileExists(this->theDefaultDirectories.toolCsvDirectory)) {
            myRuntimeInfo->out << "Create csvFiles directory at path: " << this->theDefaultDirectories.toolCsvDirectory << std::endl;
            CreateDirectory(this->theDefaultDirectories.toolCsvDirectory);
        }
        if (!fileExists(this->theDefaultDirectories.toolPltDirectory)) {
            myRuntimeInfo->out << "Create pltFiles directory at path: " << this->theDefaultDirectories.toolPltDirectory << std::endl;
            CreateDirectory(this->theDefaultDirectories.toolPltDirectory);
        }
    }
    //HTML-report
    if (this->theOutputSettings.writeHtml) {
        if (!fileExists(this->theOutputSettings.theHtmlReportDirectory)) {
            myRuntimeInfo->warn << "The given desired output directory for HTML-report '" << this->theOutputSettings.theHtmlReportDirectory << "' does not exist." << std::endl;
            myRuntimeInfo->out << "Create folder for HTML-report in given path: " << this->theOutputSettings.theHtmlReportDirectory << std::endl;
            CreateDirectory(this->theOutputSettings.theHtmlReportDirectory);
        }
    }
    //TeX-report
    if (this->theOutputSettings.writeTex) {
        if (!fileExists(this->theOutputSettings.theTexReportDirectory)) {
            myRuntimeInfo->warn << "The given desired output directory for TeX-report '" << this->theOutputSettings.theTexReportDirectory << "' does not exist." << std::endl;
            myRuntimeInfo->out << "Create folder for TeX-report in given path: " << this->theOutputSettings.theTexReportDirectory << std::endl;
            CreateDirectory(this->theOutputSettings.theTexReportDirectory);
        }
    }
}

void output::generatePlot() {
    if (this->theOutputSettings.useGnuplot) {
        myRuntimeInfo->debug << "    ... generatePlotData" << std::endl;
        this->generatePlotData();
        myRuntimeInfo->debug << "    ... generatePlotScripts" << std::endl;
        this->generatePlotScripts();
    }
    myRuntimeInfo->debug << "    ... createAndConvertPlot" << std::endl;
    this->createAndConvertPlot();
}

void output::copyPlottingFiles() {
    myRuntimeInfo->debug << "Copy plotting files from " << theDefaultDirectories.toolCsvDirectory << " to " << theDefaultDirectories.csvDirectory << std::endl;
    copyFiles(theDefaultDirectories.toolCsvDirectory, theDefaultDirectories.csvDirectory);
    myRuntimeInfo->debug << "Copy plotting files from " << theDefaultDirectories.toolPltDirectory << " to " << theDefaultDirectories.pltDirectory << std::endl;
    copyFiles(theDefaultDirectories.toolPltDirectory, theDefaultDirectories.pltDirectory);
}

void output::deletePlottingFiles() {
    myRuntimeInfo->debug << "Delete " << theDefaultDirectories.toolCsvDirectory << std::endl;
    deleteDirectory(theDefaultDirectories.toolCsvDirectory, false);
    myRuntimeInfo->debug << "Delete " << theDefaultDirectories.toolPltDirectory << std::endl;
    deleteDirectory(theDefaultDirectories.toolPltDirectory, false);
}

void output::createAndConvertPlot() {
    if (this->theOutputSettings.useGnuplot) {
        for (const plotArguments& aPlot : plots) {
            myRuntimeInfo->info << "... " << aPlot.name << std::endl;
            this->runGnuplot(aPlot);
            if (this->theOutputSettings.writeTex) {
                this->convertSVG2PDF(aPlot.name);
            }
        }
    } else {
        this->generateSvgPlot();
        if (this->theOutputSettings.writeTex) {
            for (const plotArguments& aPlot : plots) {
                this->convertSVG2PDF(aPlot.name);
            }
        }
    }
}

void output::runGnuplot(plotArguments gnuplotArgs) {
    /* add output name if given */
    if (!gnuplotArgs.outputName.empty()) {
        gnuplotArgs.additionalArguments += "set output '" + this->theOutputSettings.thePlotDirectory
                                        + gnuplotArgs.outputName + ".svg'";
        myRuntimeInfo->debug << "gnuplot set output: " << gnuplotArgs.additionalArguments << std::endl;
    }
    /* Plot */
    if (fileExists(gnuplotArgs.plotScriptFilename)) {
        std::string plotCommand = "\"" + this->theOutputSettings.gnuplotPath + "\""
                                  + " -e \"" + gnuplotArgs.additionalArguments + "\""
                                  + " " + gnuplotArgs.plotScriptFilename;
        handleChildProcess(plotCommand, "");
    } else {
        myRuntimeInfo->err << "Plot-File '" << gnuplotArgs.plotScriptFilename << "' does not exist. Cannot plot." << std::endl;
    }
}

void output::convertAllSVG2PDF(std::string aDirectory) { // cppcheck-suppress unusedFunction
    if (aDirectory == "GLOBAL_PLOTDIR") {
        aDirectory = this->theOutputSettings.thePlotDirectory;
    }
    std::vector<std::string> allSVGinPlotDIR(listDirFiles(aDirectory, "*.svg"));
    for (const std::string& svgFile : allSVGinPlotDIR) {
        this->convertSVG2PDF(svgFile);
    }
}

void output::convertSVG2PDF(const std::string& aFile) {
    myRuntimeInfo->debug << "    - Convert " << aFile << " to PDF..." << std::endl;
    std::string filename(aFile);
    filename = this->theOutputSettings.thePlotDirectory + filename.substr(0, filename.find_last_of("."));
    #ifdef _WIN32
    std::string convertCommand = this->theOutputSettings.inkscapePath + " -D -z --file=" + filename + ".svg " + " --export-pdf=" + filename + ".pdf --export-latex";
    #else
    std::string convertCommand = this->theOutputSettings.inkscapePath + " -D  --export-type=pdf --export-latex " + filename + ".svg "+">/dev/null 2>&1";
    //hide command line outputs & errors of inkscape appimage
    #endif // _WIN32
    handleChildProcess(convertCommand,  "");
}

void output::generateHtmlReport() {
    std::ofstream report;
    std::string reportFile(this->theOutputSettings.theHtmlReportDirectory + this->mProgramName + "_report.html");
    report.open(reportFile.c_str());
    if (!report) {
        myRuntimeInfo->err << this->mProgramName + "_report.html could not be opened. Abort program." << std::endl;
        exit(1);
    } else {
        report << (this->generateHtmlReportHeader()).str();
        report << (this->generateHtmlReportBody()).str();
        report << (this->generateHtmlReportFooter()).str();
        report.close();
    }
}

std::stringstream output::generateHtmlReportHeader() {
    std::stringstream report;
    /* Header of html report. Do not change! */
    report << "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"" << std::endl;
    report << "   \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">" << std::endl;
    report << "<html xmlns=\"http://www.w3.org/1999/xhtml\">" << std::endl;
    report << "<head>" << std::endl;
    report << "<meta http-equiv=\"content-type\" content=\"text/html; charset=ISO-8859-1\"/>" << std::endl;
    report << "<title>" << this->mProgramName << " " << this->theAircraft << "</title>" << std::endl;
    report << "<style type=\"text/css\">" << std::endl;
    report << "body,td,a,p{font-family:Times,sans-serif;font-size:12pt}" << std::endl;
    report << "h1{font-family:Times,sans-serif;font-size:14pt}" << std::endl;
    report << "h2{font-family:Times,sans-serif;font-size:12pt}" << std::endl;
    report << "body{color:#333;margin:15px;background:#fff;}" << std::endl;
    report << "table,img{border:0;max-width:100%}" << std::endl;
    report << "td,th{padding:5px;margin:5px}" << std::endl;
    report << ".ceras{width:100%}" << std::endl;
    report << ".ceras2Col{width:200%}" << std::endl;
    report << "#wrapper{width:100%;margin-bottom:1em;float:left}" << std::endl;
    report << "#leftCol{width:47.5%;float:left;margin-bottom:2em;}" << std::endl;
    report << "#rightCol{width:47.5%;margin-left:52.5%;margin-bottom:2em;}" << std::endl;
    report << "</style>" << std::endl;
    report << "</head>" << std::endl;
    /* End of header */
    /* Begin of body */
    report << "<body>" << std::endl;
    report << "<h1>Report of " << this->mProgramName << ": " << this->theAircraft << "</h1>" << std::endl;
    report << "<!-- Begin CeRAS Report -->" << std::endl;
    report << "<!-- ################################################################################################### -->" << std::endl;
    report << "<div id=\"wrapper\">" << std::endl;
    return report;
}

std::stringstream output::generateHtmlReportFooter() {
    std::stringstream report;
    /* End of body */
    report << "</div>" << std::endl;
    report << "<font size=\"2\">All outputs of the program '" << this->mProgramName << "' were created with version " << this->programVersion << std::endl;
    report << "</body>" << std::endl;
    /* Footer of html report. Do not change! */
    report << "</html>\r";
    return report;
}

void output::generateTexReport() {
    std::ofstream report;
    std::string reportFile(this->theOutputSettings.theTexReportDirectory + this->mProgramName + "_report.tex");
    report.open(reportFile.c_str());
    if (!report) {
        myRuntimeInfo->err << this->mProgramName + "_report.html could not be opened. Abort program." << std::endl;
        exit(1);
    } else {
        report << (this->generateTexReportBody()).str();
        report.close();
    }
}

void output::generateInfoFiles() {
    std::ofstream plot;
    std::string plotDataFile(mProgramName + "_info.txt");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << mProgramName + "_info.txt konnte nicht geoeffnet werden!" << std::endl;
        exit(1);
    } else {
        plot << "Input:" << std::endl;
        plot << "==================================================================================" << std::endl;
        for (size_t i = 0; i < node::input.size(); i++) {
            plot << node::input.at(i) << std::endl;
        }
        plot << std::endl;
        plot << "Output:" << std::endl;
        plot << "==================================================================================" << std::endl;
        for (size_t i = 0; i < node::output.size(); i++) {
            plot << node::output.at(i) << std::endl;
        }
        plot.close();
    }
}

void output::generateInputOutputXMLFile(std::vector<std::string> variablesVector, std::string direction) {
    std::sort(variablesVector.begin(), variablesVector.end());
    variablesVector.erase(std::unique(variablesVector.begin(), variablesVector.end()), variablesVector.end());
    std::ofstream plot;
    std::string plotDataFile;
    std::string filePostfix;
    if (direction.compare("input") == 0) {
        filePostfix = "_inputs.xml";
    } else if (direction.compare("output") == 0) {
        filePostfix = "_outputs.xml";
    }
    plotDataFile = mProgramName + filePostfix;
    plot.open(plotDataFile.c_str());
    std::string programInputTag("ExchangeData");
    plot << "<" << programInputTag <<"></" << programInputTag << ">";
    plot.close();
    const node& inXML = aixml::openDocument(mProgramName + filePostfix);
    std::string xmlFile = variablesVector.at(0).substr(0, variablesVector.at(0).rfind(".xml/")) + ".xml";
    node currentInputNode = aixml::openDocument(xmlFile);
    for (uint32_t i(0); i< variablesVector.size(); i++) {
        std::string polarFileString = "_polar.xml";     // Special case: skip _polar xml file
        std::size_t found = (variablesVector.at(i).substr(0, variablesVector.at(i).rfind(".xml/")) + ".xml").find(polarFileString);
        if (found != std::string::npos) {
            continue;
        }
        // If the read file is the same as from the previous iteration, use old node. Otherwise open new node.
        if (xmlFile != variablesVector.at(i).substr(0, variablesVector.at(i).rfind(".xml/")) + ".xml") {
            xmlFile = variablesVector.at(i).substr(0, variablesVector.at(i).rfind(".xml/")) + ".xml";
            myRuntimeInfo->debug << ".. read " << xmlFile << std::endl;
            currentInputNode = aixml::openDocument(xmlFile);
        }
        std::string ithXmlFile = (variablesVector.at(i).substr(0, variablesVector.at(i).find(".xml/")));
        ithXmlFile = ithXmlFile + ".xml";
        std::string resPath = variablesVector.at(i).substr(variablesVector.at(i).find(ithXmlFile)+ithXmlFile.size() , variablesVector.at(i).size());
        std::vector<std::string> tmpElements;
        std::vector<std::string> tmpPathElements = tokenize(resPath, "/");
        for (uint16_t k(0); k < tmpPathElements.size(); k++) {
            tmpElements.push_back(tmpPathElements.at(k));
        }
        std::vector<node> elementNodes;
        std::string accumulatedPath = "";
        std::vector<std::string> accumulatedPaths;
        std::string currentElement;
        std::string parentElement = programInputTag;
        for (uint16_t j = 0; j < tmpElements.size(); j++) {
            std::string accumulatedPath_old = accumulatedPath;
            accumulatedPath = accumulatedPath + "/" + tmpElements.at(j);
            /*TODO(aigner#1#05/10/21): CAUTION! The following lines are only a quick fix.
            Tools that change the aircraft XML file before the toolOutput, could cause problems, because initially read xml elements could have been deleted.*/
            if (!currentInputNode.find(accumulatedPath, 1)) {
                std::vector<std::string> tmpElement_tokenized = tokenize(tmpElements.at(j), "@");
                std::string elementName = tmpElement_tokenized.at(0);
                node& childNode = currentInputNode.at(accumulatedPath_old).appendChild(elementName, true);
                if (tmpElement_tokenized.size() > 1) {
                    std::vector<std::string> attribute_tokenized = tokenize(tmpElement_tokenized.at(1), "=");
                    std::string attributeName = attribute_tokenized.at(0);
                    std::string attributeValue = attribute_tokenized.at(1);
                    childNode.addAttrib(attributeName, attributeValue);
                }
            }
            elementNodes.push_back(currentInputNode.at(accumulatedPath));
            elementNodes.back().deleteChildren();
            std::map<std::string, std::string> currentAttribMap = elementNodes.back().getAttribMap();
            std::map<std::string, std::string>::iterator it;
            for (it = currentAttribMap.begin(); it != currentAttribMap.end(); ++it) {
                if (!compareStrings(it->first, "ID", true) &&
                    !compareStrings(it->first, "uID", true) &&
                    !compareStrings(it->first, "Name", true) &&
                    !compareStrings(it->first, "name", true)) {
                    elementNodes.back().deleteAttrib(it->first);
                } else {
                    myRuntimeInfo->debug << "... Keeping " << it->first << ": " << it->second << std::endl;
                }
            }
            if (currentAttribMap.count("Name")) {
                accumulatedPath = accumulatedPath + "@Name=" + currentAttribMap.at("Name");
                myRuntimeInfo->debug << "... Found Name tag at: " << accumulatedPath << std::endl;
            }
            accumulatedPaths.push_back(accumulatedPath);
            currentElement = programInputTag + accumulatedPaths.at(j);
            if (!inXML.find(currentElement, 1) && inXML.find(parentElement, 1)) {
                if (j == (tmpElements.size()-1)) {
                    inXML.at(parentElement).appendChild(elementNodes.at(j));
                    parentElement = programInputTag;
                } else {
                    inXML.at(parentElement).appendChild(elementNodes.at(j));
                    parentElement = currentElement;
                }
            } else {
                parentElement = currentElement;
            }
        }
    }
    aixml::saveDocument(inXML);
    aixml::closeDocument(inXML);
}

void output::generateGuiSettingsFile() {
    std::ofstream plot;
    std::string plotDataFile(mProgramName + "_guiSettings.xml");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << mProgramName + "_guiSettings.xml konnte nicht geoeffnet werden!" << std::endl;
        exit(1);
    } else {
        plot << "<?xml version=\"1.0\" encoding=\"utf-8\" standalone=\"no\"?>" << std::endl;
        plot << "<!DOCTYPE ModulSettings []>" << std::endl;
        plot << "<ModulSettings>" << std::endl;
        plot << "    <TapSetup>" << std::endl;
        plot << "        <FrontPage TapName=\"Beschreibung\"><![CDATA[" << mProgramName << "]]></FrontPage>" << std::endl;
        plot << "        <Plot TapName=\"Report\"><![CDATA[" << "reportHTML" << FILESEPERATOR << mProgramName <<  "_report.html]]></Plot>" << std::endl;
        plot << "        <Values TapName=\"Werte\">" << std::endl;
        for (size_t i(0); i < node::input.size(); i++) {
            plot << "            <Input>";
            plot << node::input.at(i) << "</Input>" << std::endl;
        }
        for (size_t i(0); i < node::output.size(); i++) {
            plot << "            <Output>";
            plot << node::output.at(i) << "</Output>" << std::endl;
        }
        plot << "        </Values>" << std::endl;
        plot << "    </TapSetup>" << std::endl;
        plot << "</ModulSettings>";
    }
    plot.close();
}

// cppcheck-suppress unusedFunction
void output::generatePlotScriptHeader(std::ofstream &script, const std::string& format, const std::string& delimiter, const bool& xtraLineStyles, const bool& doubleSize) {
    // Berechnung Skalierungsfaktor
    double width(atof((format.substr(0, format.find(':'))).c_str()));
    double height(atof((format.substr(format.find(':') + 1, format.size() - 1)).c_str()));
    double factor(width / height);
    double scaleFactor(1.);
    if (doubleSize) {
        scaleFactor *= 2.;
    }
    // Schreiben des Headers
    script << "######## Plotscript-Header ########" << std::endl;
    script << "reset" << std::endl;
    script << "set terminal svg size " << scaleFactor*columnWidth * 1.3 << "," << scaleFactor*columnWidth / factor * 1.3 << " enhanced font \"Times, 16\"" << std::endl;
    script << std::endl;
    if (delimiter != "none") {
        script << "set datafile separator \"" << delimiter << "\"" << std::endl;
        script << std::endl;
    }
    script << std::endl;
    script << "# Dashtypes can be set by value (1-5), string \".\", \"-\", \"_\", or pattern (s1,e1,s2,e2,s3,e3,s4,e4) "
           << "with 1-4 numerical pairs of <solid length>s and <empty space length>e" << std::endl;
    script << "#     - black/grey -" << std::endl;
    script << "set style line  1  dt  1     lc rgb \"black\"      lw 2 pt 1  ps 0.75#black" << std::endl;
    script << "set style line  2  dt  2     lc rgb \"#696969\"    lw 2 pt 1  ps 0.75#DimGrey" << std::endl;
    script << "set style line  3  dt  3     lc rgb \"#BEBEBE\"    lw 2 pt 1  ps 0.75#Grey" << std::endl;
    script << "set style line  4  dt  4     lc rgb \"#D3D3D3\"    lw 2 pt 1  ps 0.75#LightGrey" << std::endl;
    script << "set style line  5  dt  5     lc rgb \"#708090\"    lw 2 pt 1  ps 0.75#SlateGrey" << std::endl;
    script << "set style line  6  dt \"_\"   lc rgb \"#9FB6CD\"    lw 2 pt 5  ps 0.75#SlateGrey3" << std::endl;
    script << "set style line  7  dt \"-\"   lc rgb \"#C6E2FF\"    lw 2 pt 6  ps 0.75#SlateGrey1" << std::endl;
    script << "set style line  8  dt \"._\"  lc rgb \"#528b8b\"    lw 2 pt 2  ps 0.75#DarkSlateGray4" << std::endl;
    script << "set style line  9  dt \"..-\"  lc rgb \"#79CDCD\"    lw 2 pt 3  ps 0.75#DarkSlateGray3" << std::endl;
    if (xtraLineStyles) {
        script << std::endl;
        script << "#     - blue -" << std::endl;
        script << "set style line 11  dt  1     lc rgb \"#104E8B\"    lw 2 pt 8  ps 0.75#DogerBlue4" << std::endl;
        script << "set style line 12  dt  2     lc rgb \"#473C8B\"    lw 2 pt 8  ps 0.75#SlateBlue4" << std::endl;
        script << "set style line 13  dt  3     lc rgb \"#3A5FCD\"    lw 2 pt 8  ps 0.75#RoyalBlue3" << std::endl;
        script << "set style line 14  dt  4     lc rgb \"#0000FF\"    lw 2 pt 8  ps 0.75#Blue" << std::endl;
        script << "set style line 15  dt  5     lc rgb \"#4682B4\"    lw 2 pt 8  ps 0.75#SteelBlue" << std::endl;
        script << "set style line 16  dt \"_\"   lc rgb \"#4F94CD\"    lw 2 pt 8  ps 0.75#SteelBlue3" << std::endl;
        script << "set style line 17  dt \"-\"   lc rgb \"#6CA6CD\"    lw 2 pt 8  ps 0.75#SkyBlue3" << std::endl;
        script << "set style line 18  dt \"._\"  lc rgb \"#63B8FF\"    lw 2 pt 8  ps 0.75#SteelBlue1" << std::endl;
        script << "set style line 19  dt \"..-\"  lc rgb \"#CAE1FF\"    lw 2 pt 8  ps 0.75#LightSteelBlue1" << std::endl;
        script << std::endl;
        script << "#     - red -" << std::endl;
        script << "set style line 21  dt  1     lc rgb \"#B22222\"    lw 2 pt 7  ps 0.75#firebrick" << std::endl;
        script << "set style line 22  dt  2     lc rgb \"#CD0000\"    lw 2 pt 7  ps 0.75#red3" << std::endl;
        script << "set style line 23  dt  3     lc rgb \"#FF0000\"    lw 2 pt 7  ps 0.75#red" << std::endl;
        script << "set style line 24  dt  4     lc rgb \"#FF3030\"    lw 2 pt 7  ps 0.75#firebrick1" << std::endl;
        script << "set style line 25  dt  5     lc rgb \"#FF4500\"    lw 2 pt 7  ps 0.75#orangered" << std::endl;
        script << "set style line 26  dt \"_\"   lc rgb \"#FF6347\"    lw 2 pt 7  ps 0.75#tomato" << std::endl;
        script << "set style line 27  dt \"-\"   lc rgb \"#FF6A6A\"    lw 2 pt 7  ps 0.75#IndianRed1" << std::endl;
    }
    script << std::endl;
    script << "set key font \"Times, 12\" samplen 1 width -2" << std::endl;
    script << "###################################" << std::endl;
    script << std::endl;
}

output::plotArguments::plotArguments(const std::string& plotName, const std::string& plotScriptName, const std::string& outName, const std::string& addArgs)
    :
    name(plotName),
    additionalArguments(addArgs) {
    if (outName == "DEFAULT") {
        this->outputName = this->name;
    } else {
        this->outputName = outName;
    }
    if (plotScriptName == "DEFAULT") {
        this->plotScriptFilename = this->name;
    } else {
        this->plotScriptFilename = plotScriptName;
    }
    if (!(this->plotScriptFilename.substr(this->plotScriptFilename.find_last_of(".") + 1) == "plt")) {
        this->plotScriptFilename +=  ".plt";
    }
}
